/* LANGUAGE:    C        VAX VMS COMPILER VERSION 2.0-003


 PACKAGE NAME:          LEVAL

 PACKAGE ROUTINES:      eval_step()
			split_and_delay_and_transform()
			set_as_tr()
			set_bex()
			init_insearch()
                        mod_tran()
                        eval_total()
                        rt_organize()
                        set_delay()
                        val_d()
                        del_insearch()
                        main_level()
                        level_step()
                        del_tst()
                        main_loop()
                        loop_step()
                        main_atpg()
                        atpg_step()
			atpg_levelizing()
			atpg_set_lev()
			cfunc()
			in_func()
			in_modc()
                        val_dc()
			ret_ptr_exp_ex()
			cre_exp_el()

 PACKAGE FUNCTION:      Evaluate statistical data.
                        Look for 0-Delay loop.
                        Perform ATPG levelizing.
                        Perform circuit levelizing for simulation.
                        Perform global controls on circuit's connections.
                        Modify circuit layout for simulation needs.

 PACKAGE AUTHOR(s):     Daniela Magnani

 PACKAGE CREATION DATE: 21/08/1986

**********************************************************************/

# include <stdio.h>
# include "ldef.h"
# include "lstruct.h"
# include <math.h>

/***************************************************************************/
/*                                                                         */
/*                          VARIABLES DECLARATION                          */
/*                                                                         */
/***************************************************************************/

extern EHEADER  h_exe;
HDELAY_PTR      del_headhash[DIM_DELHASH]; /* delay hash table */
INIT_PTR        init_headhash[DIM_INITHASH]; /* initialization hash table */
HDELAY_PTR	first_d = NULL;
HDELAY_PTR	last_d = NULL;
HMC_PTR		mod_headhash[DIM_MODCHASH];  /* C modules hash table */
HFC_PTR		fun_headhash[DIM_FUNCHASH];  /* C functions hash table */
HMC_PTR		first_m = NULL;
HMC_PTR		last_m = NULL;
HFC_PTR		first_f = NULL;
HFC_PTR		last_f = NULL;

static unsigned long    pres_comp;
static unsigned long    tot_con = 0;
static unsigned long    c_del_loop = 0;
static unsigned long    c_atpg_loop = 0;
static unsigned char    loop_flag = FALSE;
static unsigned char    ill_tran = FALSE;

static unsigned char	init_flag = FALSE; /* TRUE if there are initialization 
					      values */
static unsigned char	func_flag = FALSE; /* TRUE if there are C-functions */
static unsigned char	del_ovf = FALSE;   /* to signal delay overflow */

static unsigned long	new_index;	/* number of elements in the 
					   expansion list */
static AUX_ST_PTR	first;		/* start pointer to the expansion
					   list */
static HDELAY_PTR	p_0_d;		/* pointer to the delay 0,0 */

/* debug variables */
STAT_PTR *deb;
STAT_EV_PTR *deb_ev;


/*************************************************************************

   ROUTINE:     eval_step

   FUNCTION:    main

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

eval_step()

{
/* external variables */

	extern	unsigned char	ch_delay;

/* local variables */ 

	unsigned long	i; 

/* code */ 

	for (i = 0; i < DIM_DELHASH; i++) del_headhash[i] = NULL;

        split_and_delay_and_transform();
        if (! ill_tran)
        {
                main_loop();
		if (c_del_loop > 0)
		{
			warn2d(c_del_loop, "0-Delay loop");
			if (! ch_delay) loop_flag = TRUE;
		}

                if (!loop_flag)
                {
                        
			main_level();

                        eval_total();

                        main_atpg();
                        
                        if (c_atpg_loop > 0) warn2d(c_atpg_loop, "ATPG loop");

                        if ((h_exe.n_proc) ||
                            (h_exe.n_mem > 0) || (h_exe.n_rtas > 0) || 
                            (h_exe.n_rtcn > 0) || (h_exe.n_pb > 0) ||
                            (c_atpg_loop > 0)) h_exe.f_atpg = FALSE;
			else
				
				atpg_levelizing();
                }
        }

        if (! h_exe.f_sim) warn("description invalid for simulator");
        if (! h_exe.f_atpg) warn("description invalid for ATPG");

        return(loop_flag | ill_tran);
}


/*************************************************************************

   ROUTINE:     split_and_delay_and_transform()

   FUNCTION:    A) Generate an auxiliary  AS_TR between a carrier (or gate 
		   integrated with the carrier) and a block (RT_LOAD or RT_CNT)
		   triggered by the carrier.
		   If the carrier is both trigger0 both trigger1 both trigger
		   (change) three different AS_TR are made.
		B) Generate an auxiliary AS_TR_CN between a carrier (or gate
		   integrated with the carrier) and a block RT_CNT controlled
		   by the carrier.
		   One only AS_TR_CN is made.
                C) Create an element BIT_EXtraction between a multidimensional
                   carrier and a gate having the carrier in input.
		D) Create an element BIT_EXtraction between a multidimensional 
		   carrier and a transistor's gate.
		E) Integrate delays
                F) Carriers and Gates initialized :
                   insert initialization value into the hash initialization
		   table.
		G) NODES PI:
		   Set default size value.
		   Modify connectivity to transistor
                H) Register:
                   when a register has into its fanin both RT_LD both RT_AS
                   modify the RT_LD elements strength
		I) Constants of wire type:
		   Modify connectivity to transistor
		L) C-modules and functions tables management
		M) Gates:
		   Control fanin and fanout connections

   AUTHOR(s):   Daniela Magnani

   DATE:        08/10/1986

   NOTES: 

   METHOD:      The topological elements created in this linker step are 
		inserted into the expansion list pointed by the AUX_ST_PTR
		first.
		The new elements are of the following types:
		 - dummy BIT_EX
		 - dummy AS_TR
		 - dummy AS_TR_CN
		 - PI constant (of D_CONST type)

**********************************************************************/

split_and_delay_and_transform()

{
/* external functions */

	HDELAY_PTR	del_insearch();
	HDELAY_PTR	set_delay();
        CONN_PTR        search_ty();

/* local variables */
        
        unsigned long   i;
        unsigned long   ty;
        CONN_PTR        l_f;
	CONN_PTR	list;

/* code */

	/* initialization */
	first = NULL;
        new_index = h_exe.n_top;
	p_0_d = del_insearch(0, DEL_IMM, 0, DEL_IMM);

	for (i = 0; i < h_exe.n_top; i++)
	{
	   switch (H_S[i]->code)
	   {
            case LOCV :
		/* initialization value */
		if (H_S[i]->C.init_el != NULL)
			init_insearch(i, H_S[i]->C.init_el);
                break;

            case NOD :
            case WIR :
            case REG :
		ty = H_S[i]->code;
		/* delay */
		if (H_S[i]->attr != PI)
			H_E[i]->pdelay = set_delay(H_S[i]->C.delay.del_ptr);
		else
		if (ty == NOD)
		{
		    /* PI NODE */
		    H_S[i]->C.size = DEF_SIZE;
 		    if (H_S[i]->C.l_conn != NULL) mod_tran(i, H_S[i]->C.l_conn);
		}

		if (ty == REG && (list = H_S[i]->C.l_fanin) != NULL &&
	            search_ty(RT_LD, list) != NULL && 
		    search_ty(RT_AS, list) != NULL)
   		{
                   /* register has into fanin list both RT_AS both RT_LD */
                   while (list != NULL)
                   {
                        if (list->type == RT_LD)
                        {
                                H_S[list->signal]->A.str_0 = 1;
                                H_S[list->signal]->A.str_1 = 1;
                        }
                        list = list->c_next;
                   }
   		}

		/* initialization value */
		if (H_S[i]->C.init_el != NULL)
			init_insearch(i, H_S[i]->C.init_el);

		/* AS_TR or AS_TR_CN insertion */
		if (H_S[i]->function_flag != F_NO_FUNCTION)
	    	{
		   /* The element i may be TRIGGER_0 or TRIGGER_1 or CONTROL 
		      or CHANGE */
                   for (l_f = H_S[i]->C.l_fanout; l_f != NULL;l_f = l_f->c_next)
		   {
                     if ((l_f->type == RT_LD || l_f->type == RT_CNT ||
                          l_f->type == AS_TR || l_f->type == PROC) && 
                        (l_f->CC.second == C_TR_0 || l_f->CC.second == C_TR_1 ||
			 l_f->CC.second == C_CONTROL)) set_as_tr(i, ty, l_f);
		   }
		}
		/* bit extraction */
	        /* A BIT_EX is inserted between a carrier multi and a GATE */
		if (H_S[i]->C.dim > 1)
		{
                   for (l_f = H_S[i]->C.l_fanout; l_f !=NULL; l_f = l_f->c_next)
                   {
                        if (l_f->type == GAT || l_f->type == TRAN)
				set_bex(i, l_f);
		   }
		}
		break;
            case GAT :
		/* delay */
		if (H_S[i]->G.delay.del_ptr == NULL)
			H_E[i]->pdelay = p_0_d;
		else
			H_E[i]->pdelay = set_delay(H_S[i]->G.delay.del_ptr);

		/* initialization value */
		if (H_S[i]->G.init_el != NULL)
			init_insearch(i, H_S[i]->G.init_el);

		/* AS_TR or AS_TR_CN insertion */
		ty = H_S[i]->code;
		if (H_S[i]->function_flag != F_NO_FUNCTION)
	    	{
		   /* The element i may be TRIGGER_0 or TRIGGER_1 or CONTROL 
		      or CHANGE */
                   for (l_f = H_S[i]->G.l_fanout; l_f != NULL;l_f = l_f->c_next)
		   {
                     if ((l_f->type == RT_LD || l_f->type == RT_CNT ||
                          l_f->type == AS_TR || l_f->type == PROC) && 
                        (l_f->CC.second == C_TR_0 || l_f->CC.second == C_TR_1 ||
			 l_f->CC.second == C_CONTROL)) set_as_tr(i, ty, l_f);
		   }
		}
 		break;
            case RT_AS:
            case RT_LD :        
		H_E[i]->pdelay = set_delay(H_S[i]->A.delay.del_ptr);

	        if (H_S[i]->A.l_code != NULL) cfunc(i, H_S[i]->A.l_code);

 	        if (H_S[i]->code == RT_LD)
	        {
	           /* scan polish in order to find TIME */
	           for (l_f = H_S[i]->A.l_code; l_f != NULL; l_f = l_f->c_next)
	           {
		       if (l_f->type == TIME) l_f->CT.del_p.delh_ptr = 
				set_delay(l_f->CT.del_p.del_ptr);
	           }
                }
		
		break;
            case MEM :
		H_E[i]->pdelay = set_delay(H_S[i]->M.delay.del_ptr);
		break;
	    case RT_CN :
	    case RT_CNT :
	       if (H_S[i]->O.l_code != NULL) cfunc(i, H_S[i]->O.l_code);

	       if (H_S[i]->code == RT_CNT)
	       {
	           /* scan polish in order to find TIME */
	           for (l_f = H_S[i]->O.l_code; l_f != NULL; l_f = l_f->c_next)
	           {
		       if (l_f->type == TIME) l_f->CT.del_p.delh_ptr = 
				set_delay(l_f->CT.del_p.del_ptr);
	           }
               }
               break;
	    case CONST :
	       if (H_S[i]->I.type == C_WIRE && H_S[i]->IW.l_conn != NULL) 
                  mod_tran(i, H_S[i]->IW.l_conn);
               break;

            case PROC :
   		   /* delay */
                   H_E[i]->pdelay = p_0_d;
                   /* scan polish in order to find TIME */
	           for (l_f = H_S[i]->P.l_code; l_f != NULL; l_f = l_f->c_next)
	           {
		       if (l_f->type == TIME) l_f->CT.del_p.delh_ptr = 
				set_delay(l_f->CT.del_p.del_ptr);
	           }
                break;
            case AS_TR :
		/* delay */
                H_E[i]->pdelay = p_0_d;

                /* scan polish in order to find TIME */
	        for (l_f = H_S[i]->A.l_code; l_f != NULL; l_f = l_f->c_next)
	        {
		    if (l_f->type == TIME) l_f->CT.del_p.delh_ptr = 
				set_delay(l_f->CT.del_p.del_ptr);
	        }
                break;

	    default :
	       /* do nothing */
               break;
          }
      } 
      if (new_index > h_exe.n_top) integ_static_exp();
}


/*************************************************************************

   ROUTINE:     set_as_tr

   FUNCTION:    A) Generate an auxiliary  AS_TR between a carrier (or gate
		   integrated with the carrier) and a block (RT_LOAD, RT_CNT,
                   AS_TR or PROC)
		   triggered by the carrier.
		   If the carrier is both trigger0 both trigger1 both trigger
		   (change) three different AS_TR are made.
		B) Generate an auxiliary AS_TR_CN between a carrier (or gate
		   integrated with the carrier) and a block RT_CNT controlled
		   by the carrier.
		   One only AS_TR_CN is made.

   AUTHOR(s):   Daniela Magnani

   DATE:        08/10/1986

   NOTES: 

   METHOD: 

**********************************************************************/

set_as_tr(i, ty, l_f)
unsigned long i;	/* index of the carrier or gate */
unsigned long ty;	/* code of the element i */
CONN_PTR l_f;		/* fanout block of the carrier/gate i connecting
			   to RT_LD, RT_CNT, AS_TR or PROC*/

{
/* external functions */

        CONN_PTR        search_el();
        CONN_PTR        new_c_el();
        CONN_PTR        cancel();
        CONN_PTR        ins_last();

/* local variables */
        
        unsigned long   old_c;	
        CONN_PTR        old_conn,
        		aux_l,
			l_conn;
	unsigned char	found;
        unsigned long   bf;
        AUX_ST_PTR      aux;
        STAT_PTR        p_s;

/* code */

	/* old_c = index of the element RT_CNT, RT_LOAD, AS_TR or PROC*/
	old_c = l_f->signal;

	/* old_conn = pointer to the fanin element of the element old_c
	   RT_LD, RT_CNT, AS_TR or PROC referring to the i element */
	found = FALSE;
        switch (l_f->type)
        {
          case RT_CNT:
               old_conn = H_S[old_c]->O.l_fanin;break;
          case RT_LD:
          case AS_TR:
               old_conn = H_S[old_c]->A.l_fanin;break;
          case PROC:
               old_conn = H_S[old_c]->P.l_dfanin;break;
          default: bug ("Unexpected code found in set_as_tr");
        }
	while (old_conn != NULL && !found)
	{
		old_conn = search_el(i, old_conn);
		if (old_conn != NULL)
		{
		   if (old_conn->CC.second == l_f->CC.second)
			found = TRUE;
		   else
			old_conn = old_conn->c_next;
		}
	}
        if (old_conn == NULL) fatal("illegal connection");
			

 	/* carrier trigger di RT_LD o RT_CNT - devo inserire
	  AS_TR ausiliario tra il carrier e il blocco RT.
          ho due casi :
            a) ho gia' creato AS_TR per il bit del carrier che
               sto esaminando e per il tipo di trigger e non ne
	       devo quindi creare un altro
            b) creo nuovo AS_TR 

	  carrier control di RT_CNT o AS_TR - devo inserire blocco
	  di tipo AS_TR_CN se non e' gia' stato creato (ho un
          solo blocco anche se il carrier e' multi */

	/* scan carrier's fanout */
        bf = MAX_UNSIGNED;
        for (aux_l = (ty == GAT) ? H_S[i]->G.l_fanout
				 : H_S[i]->C.l_fanout;
	     aux_l != NULL && bf == MAX_UNSIGNED; aux_l = aux_l->c_next)
	{
        	if ((aux_l->type == AS_TR && l_f->CC.second != C_CONTROL)
		    || (aux_l->type == AS_TR_CN && l_f->CC.second == C_CONTROL))
                {
                	bf = aux_l->signal;
                        for (aux = first; (aux != NULL && aux->n_el != bf);
			     aux = aux->next);
                        if (aux != NULL)
                        {
                          p_s = aux->st;
                          if (l_f->CC.second != C_CONTROL &&
			      (p_s->A.l_code->CC.first != old_conn->CC.first ||
			       p_s->A.l_code->c_next->signal - OFFS !=
				  l_f->CC.second)) bf = MAX_UNSIGNED;
                        }
                        else
                          bf = MAX_UNSIGNED;
		}
         }

         /* fanout AS_TR o AS_TR_CN */
        l_conn = new_c_el(l_f->type,l_f->signal,0, l_f->CC.second);
	if (l_f->CC.second != C_CONTROL) 
        {
           l_conn->CC.first = 0; 
           l_conn->CC.second = C_AS_TR;
        }
	if (bf != MAX_UNSIGNED)
        {
         	/* bf indice AS_TR per il bit in esame */
                /* mantengo nel fanout del carrier una sola
                   connessione allo stesso AS_TR */
                if (ty != GAT)
                	cancel (l_f, H_S[i]->C.l_fanout);
		else
                	cancel (l_f, H_S[i]->G.l_fanout);
                                
                /* incremento fanout del AS_TR o AS_TR_CN */
		p_s->A.l_fanout = ins_last(l_conn, p_s->A.l_fanout);
	}
        else
        {
		cre_exp_el(&p_s, &bf);

                p_s->attr = DUMMY;
                p_s->path_ptr = H_S[i]->path_ptr;
                p_s->A.str_0 = DEF_STR;
                p_s->A.str_1 = DEF_STR;
                p_s->A.l_fanout = l_conn;

                /* AS_TR fanin */
                p_s->A.l_fanin = new_c_el(ty, i, 0, C_AS_TR);

		if (l_f->CC.second == C_CONTROL)
		{
                	p_s->code = AS_TR_CN;
                        p_s->A.l_code = NULL;
			if (ty == GAT)
                        	p_s->A.bit = 1;
			else
                        	p_s->A.bit = H_S[i]->C.dim;
		}
		else
		{
			p_s->code = AS_TR;
                        p_s->A.bit = 1;
                        p_s->A.type = 0;
                        /* AS_TR code */
                        l_conn = new_c_el(OPERAND, 0, old_conn->CC.first,
                                  	old_conn->CC.first);
			l_conn->c_next = new_c_el(OPERATOR,
                                  	l_f->CC.second + OFFS, 0, 0);
			l_conn->c_next->c_next =new_c_el(STOP_C, STOP_TRIG_1, 0, 0);
			p_s->A.l_code = l_conn;
		}
		/* modifico fanout del carrier */
                l_f->type = p_s->code;
		l_f->signal = bf;
                l_f->CC.second = C_AS_TR;
                l_f->CC.first = 0;
	}
	/* modifico fanin elemento RT_LOAD o RT_CNT */
        old_conn->type = p_s->code;
        old_conn->signal = bf;
        old_conn->CC.first = 0;
}


/*************************************************************************
 
   ROUTINE:     set_bex()

   FUNCTION:    A) Create an element BIT_EXtraction between a multidimensional
                   carrier and a gate having the carrier in input.
		B) Create an element BIT_EXtraction between a multidimensional 
		   carrier and a transistor's gate.

   AUTHOR(s):   Daniela Magnani

   DATE:        08/10/1986

   NOTES: 

   METHOD:

**********************************************************************/

set_bex(i, l_f)
unsigned long i;	/* index of the carrier */
CONN_PTR l_f;		/* fanout block of the carrier i connecting
			   to GATE or TRAN */

{
/* external functions */

        CONN_PTR        new_c_el();
        CONN_PTR        cancel();
        CONN_PTR        ins_last();
        CONN_PTR        search_el();

/* local variables */
        
        unsigned long   bf;
        unsigned long   old_c;
        CONN_PTR        l_conn,
			aux_l,
			old_conn;
        STAT_PTR        p_s;
        AUX_ST_PTR      aux;

/* code */

	/* old_c = index of the element GATE or TRAN */
	old_c = l_f->signal;
	/* verifico se ho gia' generato elemento BIT_EX per il bit in esame */
	bf = MAX_UNSIGNED;
        for (aux_l = H_S[i]->C.l_fanout; aux_l != l_f && bf == MAX_UNSIGNED;
	     aux_l = aux_l->c_next)
        {
		if (aux_l->type == BIT_EX)
		{
			bf = aux_l->signal;
                        for (aux = first; aux != NULL && aux->n_el != bf;
			     aux = aux->next);
                        p_s = aux->st;
                        if (p_s->A.l_code->CC.first != l_f->CC.second)
				bf = MAX_UNSIGNED;
		}
	}

        /* fanout of BIT_EX */
        l_conn = new_c_el(l_f->type, l_f->signal, l_f->CC.first, l_f->CC.first);
                                
	if (bf != MAX_UNSIGNED)
        {
        	cancel(l_f, H_S[i]->C.l_fanout);
                p_s->A.l_fanout = ins_last(l_conn, p_s->A.l_fanout);
	}
	else
	{
		/* creo elemento BIT_EX */
		cre_exp_el(&p_s, &bf);

		p_s->attr = DUMMY;
                p_s->code = BIT_EX;
                p_s->path_ptr = H_S[i]->path_ptr;

		p_s->A.bit = 1;
		p_s->A.str_0 = DEF_STR;
                p_s->A.str_1 = DEF_STR;
                p_s->A.l_fanout = l_conn;
                /* BIT_EX fanin */
		p_s->A.l_fanin = new_c_el(H_S[i]->code, i, 0, C_DATA);
                /* BIT_EX code */
                l_conn = new_c_el(OPERAND, 0, l_f->CC.second, l_f->CC.second);
		l_conn->c_next = new_c_el(STOP_C, STOP_AS, 0, 0);
		p_s->A.l_code = l_conn;
                                  
		/* modifico fanout carrier */
                l_f->type = BIT_EX;
                l_f->signal = bf;
                l_f->CC.first = 0;
                l_f->CC.second = C_DATA;
	}
        /* modifico GATE fanin or TRANSISTOR gate*/
	if (H_S[old_c]->code == TRAN)
	{
		l_conn = H_S[old_c]->T.gate;
		l_conn->type = BIT_EX;
		l_conn->signal = bf;
		l_conn->CC.first = GATE_PIN;
		l_conn->CC.second = 0;
	}
	else
	{
        	l_conn = search_el(i, H_S[old_c]->G.l_fanin);
                if (l_conn == NULL) fatal("illegal connection");
                l_conn->type = BIT_EX;
                l_conn->signal = bf;
                l_conn->CC.first = 0;
                l_conn->CC.second = 0;
	}
}


/*************************************************************************

   ROUTINE:     init_insearch

   FUNCTION:    Search a value into hash initialization table and create it
                if not found.

   AUTHOR(s):   Daniela Magnani

   DATE:        01/10/1986

   NOTES: 

   METHOD:      

**********************************************************************/

init_insearch(el, s)
unsigned long el;/* element to be inserted into initialization value fanout */
char *s;        /* initialization value */

{

/* external functions */

        char            *alloca();
        int             strcmp();
	long		hash();
	CONN_PTR	new_c_el();

/* local variables */

        INIT_PTR        p;
        CONN_PTR        l_f;
        long            ind; 
	int		i;

/* code */

	if (!init_flag) 
	{
		for (i = 0; i < DIM_INITHASH; i++) init_headhash[i] = NULL; 
		init_flag = TRUE;
	}

        for (p = init_headhash[ind = hash(s) % DIM_INITHASH]; 
                    (p != NULL) && (strcmp(s, p->val) != 0); p = p->next);
        if (p == NULL)
        {
                /* insert new element */
                p = (INIT_PTR) alloca(sizeof(*p));
                p->val = s;
                p->n_fanout = 0;
                p->l_fanout = NULL;
                /* LIFO link */
                p->next = init_headhash[ind];
                init_headhash[ind] = p;
                
                h_exe.n_init++;
        }
        p->n_fanout++;
        /* insert the element el into the initialization value fanout */
        l_f = new_c_el(0, el, 0, 0);
        l_f->c_next = p->l_fanout;
        p->l_fanout = l_f;
}


/*************************************************************************

   ROUTINE:     mod_tran

   FUNCTION:    Modify connections to transistor.

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD:      case 1: 
                        transistor T
                        drain A (PI carrier or constant)
                        source B (!PI)
                        set source = A and drain = B
                case 2:
                        transistor T
                        source A (PI carrier or constant)
                        drain B (!PI)
                        move connection B->T from connectivity list
                        to the fanin list.
                	27/3/1988 iff A if VDD or GND
                case 3 : 
                        source PI
                        drain PI
                        signals error
                        
**********************************************************************/

mod_tran(el, aux)
unsigned long el;
CONN_PTR aux;

{
/* external functions */

        CONN_PTR        search_el();
        CONN_PTR        cancel();
        CONN_PTR        ins_last();
	char		*ret_path();

/* local variables */

        char            path[PATHLENGTH + 1];
        CONN_PTR        aux1;
        unsigned long   tr;
        unsigned long   dr;
        unsigned long   sr;

/* code */
        
        for (; aux != NULL; aux = aux->c_next)
        {
                tr = aux->signal;
                dr = H_S[tr]->T.drain->signal;
                sr = H_S[tr]->T.source->signal;

                if ((H_S[sr]->attr == PI) && (H_S[dr]->attr == PI))
                {
                        warn2("Illegal transistor's connection",
                                        ret_path(tr, path));
	                h_exe.warn_cnt++;
                        ill_tran = TRUE;
                }

                else

                {
                   if (el == dr)
                   {
                        /* el is the drain of the transistor tr */
                        /* do nothing if the source is PI */
                        if (H_S[sr]->attr != PI)
                        {
                                /* modify pin type in the connection sr->tr */
                                aux1 = search_el(tr, H_S[sr]->C.l_conn);
                                aux1->CC.first = DRAIN_PIN;
                                /* modify pin type in the connection drain->tr*/
                                aux->CC.first = SOURCE_PIN;
                                /* exchange drain and source in tr */
                                aux1 = H_S[tr]->T.source;
                                H_S[tr]->T.source = H_S[tr]->T.drain;
                                H_S[tr]->T.drain = aux1;
                                H_S[tr]->T.source->CC.second = SOURCE_PIN;
                                H_S[tr]->T.drain->CC.second = DRAIN_PIN;
                                dr = H_S[tr]->T.drain->signal;
                                sr = H_S[tr]->T.source->signal;
                        }
                   }
                   if (el == sr && (el == VDD || el == GND))
                   {
                        /* do nothing if the drain is PI */
                        if (H_S[dr]->attr != PI)
                        {
                                /* move connection drain->transistor tr
                                   from l_conn to l_fanin */
                                /* cancel the connection drain->tr
                                   from the connectivity list and 
                                   insert it into the fanin list */
                                aux1 = search_el(tr, H_S[dr]->C.l_conn);
                                H_S[dr]->C.l_conn = cancel(aux1,
                                                         H_S[dr]->C.l_conn);
                                aux1->c_next = NULL;
                                H_S[dr]->C.l_fanin =ins_last(H_S[dr]->C.l_fanin,
                                                                aux1);
                        }
                   }
                }
        }
}


/*************************************************************************

   ROUTINE:     eval_total

   FUNCTION:    

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES:       

   METHOD: 

**********************************************************************/

# define FAN_OV         { \
                                warn2("Fanout overflow :", ret_path(i, path)); \
		                h_exe.warn_cnt++; \
                                h_exe.f_sim = FALSE; \
                        }

eval_total()

{
/* external functions */

	char		*ret_path();
	unsigned long	cnt_con_lst();

/* local variables */

        unsigned long   i;
        unsigned long   j;
        unsigned long   fan;
        char            path[PATHLENGTH + 1];
	unsigned long	conn_cnt;
        
/* code */

  for (i = 0; i < h_exe.n_top; i++)
  {
    switch (H_S[i]->code)
    {
       case NOD :
       	if ((i != VDD) && (i != GND))
        {
        	h_exe.n_node++;
                h_exe.max_size = MAX(h_exe.max_size, H_S[i]->C.size); 
	}
        h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_lst(H_S[i]->C.l_fanin));
        h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->C.l_fanout));
        h_exe.n_conn += (H_E[i]->n_conn = cnt_con_lst(H_S[i]->C.l_conn));
        tot_con += H_E[i]->n_fanin + H_E[i]->n_fanout + H_E[i]->n_conn;

        if (H_S[i]->attr != PO) tot_con++;

        h_exe.n_nfanin += H_E[i]->n_fanin;
        h_exe.n_nfanout += H_E[i]->n_fanout;
        h_exe.n_nconn += H_E[i]->n_conn;
        if (H_E[i]->n_fanout >= MAX_FAN) FAN_OV;
        h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
	break;
       case WIR :
	h_exe.n_wire++;
        h_exe.n_bitwire += H_S[i]->C.dim;
        h_exe.max_size = MAX(h_exe.max_size, H_S[i]->C.size); 
	h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_lst(H_S[i]->C.l_fanin));
	h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->C.l_fanout));
        tot_con += H_E[i]->n_fanin + H_E[i]->n_fanout;

        if (H_S[i]->attr != PO) tot_con++;

        h_exe.n_wfanin += H_E[i]->n_fanin;
        h_exe.n_wfanout += H_E[i]->n_fanout;
        if (H_E[i]->n_fanout >= MAX_FAN) FAN_OV;
        h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
        break;
       case LOCV :
	h_exe.n_locvar++;
        break;
      case REG :
	h_exe.n_reg++;
        h_exe.n_bitreg += H_S[i]->C.dim;
        h_exe.max_size = MAX(h_exe.max_size, H_S[i]->C.size); 
        h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_lst(H_S[i]->C.l_fanin));
        h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->C.l_fanout));

        tot_con += H_E[i]->n_fanin + H_E[i]->n_fanout;

        if (H_S[i]->attr != PO) tot_con++;

        h_exe.n_rfanin += H_E[i]->n_fanin;
        h_exe.n_rfanout += H_E[i]->n_fanout;
        if (H_E[i]->n_fanout >= MAX_FAN) FAN_OV;
        h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
        break;
      case TRAN :
        h_exe.n_tran++;
        h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->T.str); 
	break;
      case GAT :
        h_exe.n_gate++;
        h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->G.str_0);
	h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->G.str_1);
        h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_lst(H_S[i]->G.l_fanin));
        h_exe.n_gfanin += H_E[i]->n_fanin;
        h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->G.l_fanout));

	h_exe.n_gfanout += H_E[i]->n_fanout;
	tot_con += H_E[i]->n_fanout;
	if (H_E[i]->n_fanout >= MAX_FAN) FAN_OV;
	h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
	break;
      case RT_AS :
      case RT_LD :
      case BIT_EX :
        if (H_S[i]->A.l_fanin != NULL) rt_organize(i, H_S[i]->A.l_fanin);

      	h_exe.n_rtas++;
        h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->A.str_0);
        h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->A.str_1);
        h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_mod(H_S[i]->A.l_fanin));
        h_exe.n_asfanin += H_E[i]->n_fanin;
        h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->A.l_fanout));
        h_exe.n_asfanout += H_E[i]->n_fanout;
	H_E[i]->n_code = cnt_con_lst(H_S[i]->A.l_code);
        tot_con += (H_E[i]->n_fanout * 2) + H_E[i]->n_code + H_E[i]->n_fanin;
        if (H_E[i]->n_fanout>= MAX_FAN) FAN_OV;
        h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
        break;
      case AS_TR :
      case AS_TR_CN :
        if (H_S[i]->A.l_fanin != NULL) rt_organize(i, H_S[i]->A.l_fanin);

      	h_exe.n_rtcn++;
        h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->A.str_0);
        h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->A.str_1);
        h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_mod(H_S[i]->A.l_fanin));
        h_exe.n_cnfanin += H_E[i]->n_fanin;
        h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->A.l_fanout));
        h_exe.n_cnfanout += H_E[i]->n_fanout;
	H_E[i]->n_code = cnt_con_lst(H_S[i]->A.l_code);
        tot_con += H_E[i]->n_fanout + H_E[i]->n_code + H_E[i]->n_fanin;
        if (H_E[i]->n_fanout>= MAX_FAN) FAN_OV;
        h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
        break;
      case RT_CN :
      case RT_CNT :
        if (H_S[i]->O.l_fanin != NULL) rt_organize(i, H_S[i]->O.l_fanin);

        h_exe.n_rtcn++;
        h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_mod(H_S[i]->O.l_fanin));
        h_exe.n_cnfanin += H_E[i]->n_fanin;
        h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->O.l_fanout));
        h_exe.n_cnfanout += H_E[i]->n_fanout;
	H_E[i]->n_code = cnt_con_lst(H_S[i]->O.l_code);
        tot_con += H_E[i]->n_fanout + H_E[i]->n_code + H_E[i]->n_fanin;
	if (H_E[i]->n_fanout >= MAX_FAN) FAN_OV;
        h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
        break;
      case CONST :
        h_exe.n_const++;
	switch (H_S[i]->I.type)
	{
	   case C_WIRE :
	        h_exe.n_fanout += (H_E[i]->n_fanout = 
			cnt_con_lst(H_S[i]->IW.l_fanout));
	        h_exe.n_conn += (H_E[i]->n_conn = 
			cnt_con_lst(H_S[i]->IW.l_conn));
	        break;
	   case C_INT :
	   case D_CONST :
	        h_exe.n_fanout += (H_E[i]->n_fanout = 
			cnt_con_lst(H_S[i]->ID.l_fanout));
	        h_exe.n_fanin += (H_E[i]->n_fanin = 
			cnt_con_lst(H_S[i]->ID.l_fanin));
	        H_E[i]->n_code = cnt_con_lst(H_S[i]->ID.l_code);
	        break;
	   case C_CHAR :
 	        h_exe.n_fanout += (H_E[i]->n_fanout = 
			cnt_con_lst(H_S[i]->IC.l_fanout));
		break;
	}
        tot_con += H_E[i]->n_fanout + H_E[i]->n_fanin + H_E[i]->n_conn +
		   H_E[i]->n_code;
        if (H_E[i]->n_fanout >= MAX_FAN) FAN_OV;
        h_exe.max_fan = MAX(h_exe.max_fan, H_E[i]->n_fanout);
 	break;
      case PROC :
        if (H_S[i]->P.l_dfanin != NULL) rt_organize(i, H_S[i]->P.l_dfanin);
	h_exe.n_proc++;
        h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_lst(H_S[i]->P.l_dfanin));
        h_exe.n_fanin += (H_E[i]->n_tfanin = cnt_con_mod(H_S[i]->P.l_tfanin));
        h_exe.n_fanout += (H_E[i]->n_fanout = cnt_con_lst(H_S[i]->P.l_fanout));
        H_E[i]->n_conn = cnt_con_lst(H_S[i]->P.l_locvar);
        H_E[i]->n_code = cnt_con_lst(H_S[i]->P.l_code);

        tot_con += H_E[i]->n_fanin + H_E[i]->n_tfanin + H_E[i]->n_code + 1;

        h_exe.n_dfanin += H_E[i]->n_fanin;
        h_exe.n_tfanin += H_E[i]->n_tfanin;
        h_exe.n_pfanout += H_E[i]->n_fanout;
        if (H_E[i]->n_fanout >= MAX_FAN) FAN_OV;
        break;
      case MEM :
        h_exe.n_mem++;
        h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->M.str_0);
	h_exe.max_strength = MAX(h_exe.max_strength, H_S[i]->M.str_1);
	h_exe.n_fanin += (H_E[i]->n_fanin = cnt_con_lst(H_S[i]->M.l_fanin));
	/* fanout is 1 */
        h_exe.n_fanout ++;
        tot_con += 2 + H_E[i]->n_fanin + (2 * (1 << H_S[i]->M.addr_size));
	h_exe.max_fan = MAX(h_exe.max_fan, 1);
        break;
    }
  }
  h_exe.sb_dim = h_exe.n_tran * STRAN;
  h_exe.sb_dim += h_exe.n_gate * SGATE;
  h_exe.sb_dim += SDEF * (h_exe.n_node + h_exe.n_wire + h_exe.n_reg +
  		 h_exe.n_rtas + h_exe.n_rtcn + h_exe.n_const + h_exe.n_mem +
                 h_exe.n_proc + h_exe.n_locvar);
  h_exe.sb_dim += tot_con;
  if (VDD != MAX_UNSIGNED) h_exe.sb_dim += SDEF;
  if (GND != MAX_UNSIGNED) h_exe.sb_dim += SDEF;
}


/*************************************************************************

   ROUTINE:     rt_organize

   FUNCTION:    Move connections to RT assignements or RT conditions to
                the end of a fanout list

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

rt_organize(i, list)
unsigned long i;/* index of a topological element (RTAS || RTLD || RTCN ||
                                                   AS_TR || AS_TR_CN ) */ 
CONN_PTR list; /* pointer to the fanin list of the element i */

{
/* external functions */

        CONN_PTR        search_el();
        CONN_PTR        cancel();
        CONN_PTR        ins_last();

/* local variables */

        CONN_PTR        aux;
        CONN_PTR        aux1;
        unsigned long   el;

/* code */

        while (list != NULL)
        {
                el = list->signal;
                switch (H_S[el]->code)
                {
                        case NOD :
                        case WIR :
                        case REG :
                                aux1 = H_S[el]->C.l_fanout;
                                aux = search_el(i, aux1);
                                if (aux->c_next != NULL)
                                {
                                       aux1 = cancel(aux, aux1);
                                       aux->c_next = NULL;
                                       H_S[el]->C.l_fanout =ins_last(aux1, aux);
                                }
                                break;
                        case GAT :
                                aux1 = H_S[el]->G.l_fanout;
                                aux = search_el(i, aux1);
                                if (aux->c_next != NULL)
                                {
                                       aux1 = cancel(aux, aux1);
                                       aux->c_next = NULL;
                                       H_S[el]->G.l_fanout =ins_last(aux1, aux);
                                }
                                break;
                        case CONST :
			   switch (H_S[el]->I.type)
			   {
			     case C_WIRE :
                                aux1 = H_S[el]->IW.l_fanout;
                                aux = search_el(i, aux1);
                                if (aux->c_next != NULL)
                                {
                                       aux1 = cancel(aux, aux1);
                                       aux->c_next = NULL;
                                       H_S[el]->IW.l_fanout = ins_last(aux1, aux);
                                }
                                break;
			     case C_INT :
			     case D_CONST :
                                aux1 = H_S[el]->ID.l_fanout;
                                aux = search_el(i, aux1);
                                if (aux->c_next != NULL)
                                {
                                       aux1 = cancel(aux, aux1);
                                       aux->c_next = NULL;
                                       H_S[el]->ID.l_fanout = ins_last(aux1, aux);
                                }
                                break;
			     case C_CHAR :
                                aux1 = H_S[el]->IC.l_fanout;
                                aux = search_el(i, aux1);
                                if (aux->c_next != NULL)
                                {
                                       aux1 = cancel(aux, aux1);
                                       aux->c_next = NULL;
                                       H_S[el]->IC.l_fanout = ins_last(aux1, aux);
                                }
                                break;
			   }
			   break;
                        case PROC :
                        case LOCV :
                        case RT_CN :
                        case RT_CNT :
                        case AS_TR :
                        case AS_TR_CN :
                                /* do nothing */
                                break;
                        default :
                               fatal("Invalid element connected to RT element");
                                break;
                }
                list = list->c_next;
        }
}


/*************************************************************************

   ROUTINE:     set_delay

   FUNCTION:    Evaluate delays and return pointer to delay hash table

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD:

**********************************************************************/

HDELAY_PTR set_delay(p_d)
ODELAY_PTR p_d;

{
/* external functions */

        HDELAY_PTR      del_insearch();
	long		val_d();
	long		val_dc();

/* local variables */

        unsigned long   rise;
        unsigned long   fall;
	long		aux;
	
/* code */

	aux = p_d->time_unit - h_exe.time_unit;
	rise = (p_d->rise_type == DEL_IMM) ?
		      ((p_d->rise == 0) ? 0 : val_d(p_d->rise,aux)) :
				val_dc(p_d->rise, aux);
	fall = (p_d->fall_type == DEL_IMM) ?
		      ((p_d->fall == 0) ? 0 : val_d(p_d->fall,aux)) :
				val_dc(p_d->fall, aux);
        return (del_insearch(rise, p_d->rise_type, fall, p_d->fall_type));
}


/*************************************************************************

   ROUTINE:     val_d

   FUNCTION:    Control range of delay values (given as immediate)
		Delay scaling.

   AUTHOR(s):   Daniela Magnani

   DATE:        27/01/1989

   NOTES: 	t_un is the scale factor that must be applied
		del is the immediate delay value

   METHOD: 

**********************************************************************/

long val_d(del, t_un)
unsigned long del;
long t_un;
{

/* local variables */

	int	i;
	long	val;

/* code */

   if (t_un != 0)
   {
	/* scaling */
	for (i = 0; i < t_un; i++)
	{
		del *= 10;
		if (del > DELAY_MAX)
		{
		        del = DELAY_MAX;
			if (! del_ovf)
			{
				warn("delay overflow");
				h_exe.warn_cnt++;
				del_ovf = TRUE;
				break;
			}
		}
	}
   }
   else
   {
	if (del > DELAY_MAX)
	{
	        del = DELAY_MAX;
		if (! del_ovf)
		{
			warn("delay overflow");
			h_exe.warn_cnt++;
			del_ovf = TRUE;
		}
	}
   }
   h_exe.max_d = MAX(h_exe.max_d, del);
   /* min_d is surely 0 because 0 delay is inserted and 0 is the
     minumum value admitted */ 

   return(del);
}


/*************************************************************************

   ROUTINE:     del_insearch

   FUNCTION:    Insert a delay into delay hash table

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

HDELAY_PTR del_insearch(rise, r_ty, fall, f_ty)
unsigned long rise;
unsigned char r_ty;
unsigned long fall;
unsigned char f_ty;

{
/* external functions */

        char    *alloca();

/* local variables */

        long            i;
        HDELAY_PTR      p;

/* code */
        
        for (p = del_headhash[i = (rise * 10 + fall) % DIM_DELHASH];
                p != NULL; p = p->next)
        {
                if (rise == p->rise && fall == p->fall && 
		    r_ty == p->rise_type && f_ty == p->fall_type) return(p);
        }
        /* delay not found  - new delay creation */
        /* control delay amount */
        if (h_exe.n_delay == MAX_DELAY)
        {
                h_exe.f_sim = FALSE;
                warn("maximum number of delays exceeded");
		h_exe.warn_cnt++;
        }

        p = (HDELAY_PTR) alloca(sizeof(*p));
        p->rise = rise;
        p->fall = fall;
	p->rise_type = r_ty;
	p->fall_type = f_ty;
        p->n_delay = h_exe.n_delay++;
	p->insert = NULL;

	/* link FIFO to insertion list */
	if (last_d == NULL)
		first_d = last_d = p;
	else
		last_d = last_d->insert = p;
	
        /* link LIFO to hash delay table */
        p->next = del_headhash[i];
        del_headhash[i] = p;

        return(p);
}


/*************************************************************************

   ROUTINE:     main_level

   FUNCTION:    Assign a simulation level to each topological element

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD:      scan the topological elements table and call the function
                level_step for the elements not_visited :
                  a)  CARRIER PI (starting level is 1)
                  b)  CARRIER with delay != 0   
                  c)  CONST
                  d)  elements with delay != 0 (RT_AS, RT_LD, GAT, MEM)

**********************************************************************/

main_level()

{
/* external functions */

        long    del_tst();

/* local variables */

        unsigned long   i;

/* code */

   /* reset of visit flag */
   for (i = 0; i < h_exe.n_top; i++)
   {
	H_E[i]->visited = FALSE;
	H_E[i]->sim_lev = 0;
   }

   for (i = 0; i < h_exe.n_top; i++)
   {
      if (! H_E[i]->visited)
      {
         switch (H_S[i]->code)
         {
            case NOD :
            case WIR :
            case REG :
              if (H_S[i]->attr == PI) level_step(i, 1);
	      else if (del_tst(H_E[i]->pdelay) != 0) level_step(i, 0);
              break;

            case GAT :
            case RT_AS :
            case RT_LD :
            case MEM :
              if (del_tst(H_E[i]->pdelay) != 0) level_step(i, 0);
              break;

            case CONST :
              level_step(i, 0);
              break;

         }
      }
   }
}
 

/*************************************************************************

   ROUTINE:     level_step

   FUNCTION:    Set level for the element given as input.
                Scan fanout and connectivity lists assigning level until
                an element with delay != 0 is encountered

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

level_step(el, level)
unsigned long el;
unsigned long level;

{
/* external functions */

        long    	del_tst();
	char		*ret_path();

/* local variables */

        CONN_PTR        pf;
        CONN_PTR        pc;
        unsigned long   j;
        unsigned long   dr;
        unsigned long   sr;
        unsigned long   ad;
        char            path[PATHLENGTH + 1];

/* code */
   
   H_E[el]->sim_lev = level;
   H_E[el]->visited = TRUE;

   if (level > MAX_LEV)
   {
        warn2("Level overflow :", ret_path(el, path));
        h_exe.warn_cnt++;
        h_exe.f_sim = FALSE;
   }

   pc = NULL;
   switch (H_S[el]->code)
   {
      case GAT :
        pf = H_S[el]->G.l_fanout;
        break;
      case NOD :
      case WIR :
      case REG :
        pf = H_S[el]->C.l_fanout;
        pc = H_S[el]->C.l_conn;
        break;
      case RT_AS :
      case RT_LD :
      case BIT_EX :
      case AS_TR :
      case AS_TR_CN :
        pf = H_S[el]->A.l_fanout;
        break;
      case PROC :
        pf = H_S[el]->P.l_fanout;
        break;
      case RT_CN :
      case RT_CNT :
        pf = H_S[el]->O.l_fanout;
        break;
      case CONST :
	switch (H_S[el]->I.type)
	{
	   case C_WIRE :
	        pf = H_S[el]->IW.l_fanout;
	        pc = H_S[el]->IW.l_conn;
 	        break;
	   case C_INT :
	   case D_CONST :
 		pf = H_S[el]->ID.l_fanout;
 	        break;
	   case C_CHAR :
 		pf = H_S[el]->IC.l_fanout;
 	        break;
	}
	break;
       case MEM :
        pf = H_S[el]->M.l_fanout;
        break;
   }

   /* scan the fanout list */
   while (pf != NULL)
   {
      switch (H_S[j = pf->signal]->code)
      {
         case NOD :
         case WIR :
         case REG :
         case GAT :
         case RT_AS :
         case RT_LD :
         case MEM :
	    /* elements with delay */
            if (del_tst(H_E[j]->pdelay) == 0)
            {
              if (H_S[el]->code == PROC)
              {
                if (level > H_E[j]->sim_lev) 
                  level_step(j, level);
              }
              else
               if (level >= H_E[j]->sim_lev) 
                 level_step(j, level+1);
            }
            break;
         case RT_CN :
         case RT_CNT :
         case BIT_EX :
         case AS_TR :
         case AS_TR_CN :
         case PROC:
	    /* elements without delay */
            if (level >= H_E[j]->sim_lev) level_step(j, level+1);
            break;
         case TRAN :
            if (H_S[j]->T.type != _DEPLETION)
            {     
               /* DRAIN carrier */
               dr = H_S[j]->T.drain->signal;
               if ((H_S[dr]->attr != PI) && (del_tst(H_E[dr]->pdelay) == 0)
                    && (level >= H_E[dr]->sim_lev)) level_step(dr, level+1);
               /* SOURCE carrier */
               sr = H_S[j]->T.source->signal;
               if ((H_S[sr]->attr != PI) && (del_tst(H_E[sr]->pdelay) == 0)
                    && (level >= H_E[sr]->sim_lev)) level_step(sr, level+1);
            }
            break;
      }
      pf = pf->c_next;
   }
 
   /* scan the connectivity list */
   while (pc != NULL)
   {
      /* examine the adjacent carrier */
      if (H_S[pc->signal]->T.source->signal == el) 
                ad = H_S[pc->signal]->T.drain->signal;
      else
                ad = H_S[pc->signal]->T.source->signal;
      if ((H_S[ad]->attr != PI) && (del_tst(H_E[ad]->pdelay) == 0)
           && (level > H_E[ad]->sim_lev)) level_step(ad, level);
      pc = pc->c_next;
   }
}


/*************************************************************************

   ROUTINE:     del_tst

   FUNCTION:    Return 0 if the rise value or the fall value is 0

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

long del_tst(p)
HDELAY_PTR p;

{
/* local variables */

	long	a_rise;
	long	a_fall;

/* code */

	a_rise = (p->rise_type == DEL_IMM) ? p->rise : H_S[p->rise]->ID.evval;
	a_fall = (p->fall_type == DEL_IMM) ? p->fall : H_S[p->fall]->ID.evval;
        return((a_rise == 0 ) ? a_rise : a_fall);
}


/*************************************************************************

   ROUTINE:     main_loop

   FUNCTION:    Search 0-Delay loops.

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD:      scan the table of the topological elements and call the
                function loop_step for the elements not_visited
                        a) gate with delay == 0
                        b) carrier(!= PI) with delay == 0
                        c) RT assignements (triggered or not) 
                           with delay == 0
                        d) MEM with delay == 0

                Eliminate 0-Delay loop if requested [option -z]

**********************************************************************/

main_loop()

{
/* external functions */

        long    del_tst();

/* external variables */

        extern unsigned char    ch_delay;

/* local variables */

        unsigned long   i;

/* code */

        for (i = 0; i < h_exe.n_top; i++)
        {
                H_E[i]->visited = FALSE;
                H_E[i]->on_path = FALSE;
                H_E[i]->comp = UNDEFINED;
        }
        
        if (! ch_delay) init_stack();

        for (i = 0; i < h_exe.n_top; i++)
        {
                if ((H_S[i]->attr != PI) && (! H_E[i]->visited))
                {
                        pres_comp = i;
                        switch (H_S[i]->code)
                        {
                                case NOD :
                                case WIR :
                                case REG :
                                case GAT :
                                case RT_AS :
                                case RT_LD :
                                case MEM :
                                        if (del_tst(H_E[i]->pdelay) == 0)
                                                    loop_step(i, 0, UNDEFINED);
                                        break;
				default :
					/* do nothing */
					break;
                        }
                }
        }
}


/*************************************************************************

   ROUTINE:     loop_step

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

loop_step(el, level, adj_n)
unsigned long el;
unsigned long level;
unsigned long adj_n;

{
/* external functions */

        long    del_tst();

/* external variables */

        extern unsigned char    ch_delay;

/* local variables */

        CONN_PTR        pf;
        CONN_PTR        pc;
        unsigned long   j;
        unsigned long   dr;
        unsigned long   sr;
        unsigned long   ad;

/* code */

   H_E[el]->sim_lev = level;
   if (! H_E[el]->visited) H_E[el]->visited = STRONG;
   H_E[el]->on_path = TRUE;
   H_E[el]->comp = pres_comp;

   /* push element onto the stack */
   if (! ch_delay) push(el);

   pc = NULL;
   switch (H_S[el]->code)
   {
      case GAT :
        pf = H_S[el]->G.l_fanout;
        break;
      case NOD :
      case WIR :
      case REG :
        pf = H_S[el]->C.l_fanout;
        pc = H_S[el]->C.l_conn;
        break;
      case RT_AS :
      case RT_LD :
      case BIT_EX :
      case AS_TR :
      case AS_TR_CN :
        pf = H_S[el]->A.l_fanout;
        break;
      case PROC :
        pf = H_S[el]->P.l_fanout;
        break;
      case RT_CN :
      case RT_CNT :
        pf = H_S[el]->O.l_fanout;
        break;
      case MEM :
        pf = H_S[el]->M.l_fanout;
        break;
   }

   /* scan the fanout list */
   while (pf != NULL)
   {
      switch(H_S[j = pf->signal]->code)
      {
         case NOD :
         case WIR :
         case REG :
            if (del_tst(H_E[j]->pdelay) == 0)
            {
               if (! H_E[j]->visited) loop_step(j, level+1, UNDEFINED);
               else
               {
                  if ((H_E[j]->comp == pres_comp) && (H_E[j]->sim_lev <= level)
                                                  && (H_E[j]->on_path))
                  {
                        LOOP(j)
                        if (ch_delay) H_E[j]->pdelay = 
				del_insearch(1, DEL_IMM, 1, DEL_IMM);
                  }
                  else
                  {
                     if (H_E[j]->visited == WEAK)
                     {
                        H_E[j]->visited = STRONG;
                        loop_step(j, level+1, UNDEFINED);
                     }
                  }
               }
            }
            break;

         case RT_AS :
         case RT_LD :
         case MEM :
         case GAT :
            if (del_tst(H_E[j]->pdelay) == 0)
            {
               if (! H_E[j]->visited)
               {
                 if (H_S[el]->code == PROC)
                   loop_step(j, level, UNDEFINED);
                 else
                   loop_step(j, level+1, UNDEFINED);
               }
               else
               if ((H_E[j]->comp == pres_comp) && (H_E[j]->sim_lev <= level)
                                               && (H_E[j]->on_path)) 
               {
                        LOOP(j)
                        if (ch_delay) H_E[j]->pdelay = 
				del_insearch(1, DEL_IMM, 1, DEL_IMM);
               }
            }
            break;

/* elementi senza ritardo o dummy - non prevedo loop su di essi */
         case RT_CN :
         case RT_CNT :
         case BIT_EX :
         case AS_TR :
         case AS_TR_CN :
         case PROC :
            if (! H_E[j]->visited) loop_step(j, level+1, UNDEFINED);
            break;

         case TRAN :
            if (H_S[j]->T.type != _DEPLETION)
            {
                /* DRAIN carrier */
                dr = H_S[j]->T.drain->signal;
                if ((H_S[dr]->attr != PI) && (del_tst(H_E[dr]->pdelay) == 0))
                {
                   if (! H_E[dr]->visited) loop_step(dr, level+1, UNDEFINED);
                   else
                   {
                       if ((H_E[dr]->comp == pres_comp) && 
                           (H_E[dr]->sim_lev <= level) && (H_E[dr]->on_path))
                       {
                                LOOP(dr)
                                if (ch_delay) 
                                        H_E[dr]->pdelay = 
					  del_insearch(1, DEL_IMM, 1, DEL_IMM);
                       }
                       else
                       {
                          if (H_E[dr]->visited == WEAK)
                          {
                              H_E[dr]->visited = STRONG;
                              loop_step(dr, level+1, UNDEFINED);
                          }
                       }
                   }
                 }
  
                /* SOURCE carrier */
                sr = H_S[j]->T.source->signal;
                if ((H_S[sr]->attr != PI) && (del_tst(H_E[sr]->pdelay) == 0))
                {
                   if (! H_E[sr]->visited) loop_step(sr, level+1, dr);
                   else
                   {
                      if ((H_E[sr]->comp == pres_comp) &&
                          (H_E[sr]->sim_lev <= level) && (H_E[sr]->on_path))
                       {
                                LOOP(sr)
                                if (ch_delay) 
                                        H_E[sr]->pdelay = 
				         del_insearch(1, DEL_IMM, 1, DEL_IMM);
                       }
                      else
                      {
                         if (H_E[sr]->visited == WEAK)
                         {
                            H_E[sr]->visited = STRONG;
                            loop_step(sr, level+1, dr);
                         }
                      }
                   }
                 }
            }
            break;
      }
      pf = pf->c_next;
   }

   /* scan the connectivity list */
   while (pc != NULL)
   {
      /* adjacent carrier */
      if (H_S[pc->signal]->T.source->signal == el) 
                ad = H_S[pc->signal]->T.drain->signal;
      else
                ad = H_S[pc->signal]->T.source->signal;
      if ((ad != adj_n) && (H_S[ad]->attr != PI)
                        && (del_tst(H_E[ad]->pdelay) == 0))
      {
         if (! H_E[ad]->visited)
         {
            H_E[ad]->visited = WEAK;
            loop_step(ad, level, el);
         }
         else
         {
            if ((H_E[ad]->comp == pres_comp) && (H_E[ad]->sim_lev < level)
                                             && (H_E[ad]->on_path)) 
            {
                LOOP(ad)
                if (ch_delay) H_E[ad]->pdelay = 
			del_insearch(1, DEL_IMM, 1, DEL_IMM);
            }
            else
            {
               if (H_E[ad]->visited == WEAK)
               {
                   H_E[ad]->visited = STRONG;
                   loop_step(ad, level, el);
               }
            }
         }
      }
      pc = pc->c_next;
   }
   /* pop from stack */
   H_E[el]->on_path = FALSE;
   if (! ch_delay) pop();
}


/*************************************************************************

   ROUTINE:     main_atpg

   FUNCTION:    Search ATPG loops 

   AUTHOR(s):   Daniela Magnani

   DATE:        15/04/1987

   NOTES: 

   METHOD:      scan the table of the topological elements and call the
                function atpg_step for the elements not_visited
                        a) Primary Inputs
			b) Flip-Flop

**********************************************************************/

main_atpg()

{
/* external variables */

	extern	unsigned char	atpg_flag;

/* local variables */

        unsigned long   i;

/* code */

        for (i = 0; i < h_exe.n_top; i++)
        {
                H_E[i]->visited = FALSE;
                H_E[i]->on_path = FALSE;
                H_E[i]->comp = UNDEFINED;
        }

        if (atpg_flag) init_stack();

        for (i = 0; i < h_exe.n_top; i++)
        {
                if (! H_E[i]->visited &&
		    (H_S[i]->attr == PI || 
		     (H_S[i]->code == GAT && (H_S[i]->G.type >= FIRST_FF &&
			                      H_S[i]->G.type <= LAST_FF))))
		{
                        pres_comp = i;
                        atpg_step(i, 0, UNDEFINED);
                }
        }
}


/*************************************************************************

   ROUTINE:     atpg_step

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        15/04/1987

   NOTES: 

   METHOD: 

**********************************************************************/

atpg_step(el, level, adj_n)
unsigned long el;
unsigned long level;
unsigned long adj_n;

{
/* external variables */

	extern	unsigned char	atpg_flag;

/* local variables */

        CONN_PTR        pf;
        CONN_PTR        pc;
        unsigned long   j;
        unsigned long   dr;
        unsigned long   sr;
        unsigned long   ad;

/* code */

   H_E[el]->atpg_lev = level;
   if (! H_E[el]->visited) H_E[el]->visited = STRONG;
   H_E[el]->on_path = TRUE;
   H_E[el]->comp = pres_comp;

   /* push element onto the stack */
   if (atpg_flag) push(el);

   pc = NULL;
   switch (H_S[el]->code)
   {
      case GAT :
        pf = H_S[el]->G.l_fanout;
        break;
      case NOD :
      case WIR :
      case REG :
        pf = H_S[el]->C.l_fanout;
        pc = H_S[el]->C.l_conn;
        break;
      case RT_AS :
      case RT_LD :
      case BIT_EX :
      case AS_TR :
      case AS_TR_CN :
        pf = H_S[el]->A.l_fanout;
        break;
      case PROC :
        pf = H_S[el]->P.l_fanout;
        break;
      case RT_CN :
      case RT_CNT :
        pf = H_S[el]->O.l_fanout;
        break;
      case MEM :
        pf = H_S[el]->M.l_fanout;
        break;
      case CONST :
	switch (H_S[el]->I.type)
	{
	   case C_WIRE :
	        pf = H_S[el]->IW.l_fanout;
	        pc = H_S[el]->IW.l_conn;
 	        break;
	   case C_INT :
	   case D_CONST :
 		pf = H_S[el]->ID.l_fanout;
 	        break;
	   case C_CHAR :
 		pf = H_S[el]->IC.l_fanout;
 	        break;
	}
	break;
   }

   /* scan the fanout list */
   while (pf != NULL)
   {
      switch(H_S[j = pf->signal]->code)
      {
         case RT_AS :
         case RT_LD :
         case RT_CN :
         case RT_CNT :
         case BIT_EX :
         case AS_TR :
         case AS_TR_CN :
         case MEM :
         case PROC :
                if (! H_E[j]->visited) atpg_step(j, level+1, UNDEFINED);
                else
                if ((H_E[j]->comp == pres_comp) && (H_E[j]->atpg_lev <= level)
                                                && (H_E[j]->on_path)) ALOOP(j)
                break;
         case GAT :
		if (H_S[j]->G.type < FIRST_FF || H_S[j]->G.type > LAST_FF)
                {
		   if (! H_E[j]->visited) atpg_step(j, level+1, UNDEFINED);
                   else
                   if ((H_E[j]->comp == pres_comp) && (H_E[j]->atpg_lev<= level)
                                                && (H_E[j]->on_path)) ALOOP(j)
		}
                break;
         case NOD :
         case WIR :
         case REG :
               if (! H_E[j]->visited) atpg_step(j, level+1, UNDEFINED);
               else
               {
                  if ((H_E[j]->comp == pres_comp) && (H_E[j]->atpg_lev <= level)
                                                  && (H_E[j]->on_path)) ALOOP(j)
                  else
                  {
                     if (H_E[j]->visited == WEAK)
                     {
                        H_E[j]->visited = STRONG;
                        atpg_step(j, level+1, UNDEFINED);
                     }
                  }
               }
               break;
         case TRAN :
            if (H_S[j]->T.type != _DEPLETION)
            {
                /* DRAIN carrier */
                dr = H_S[j]->T.drain->signal;
                if (H_S[dr]->attr != PI)
                {
                   if (! H_E[dr]->visited) atpg_step(dr, level+1, UNDEFINED);
                   else
                   {
                       if ((H_E[dr]->comp == pres_comp) && 
                           (H_E[dr]->atpg_lev <= level) && (H_E[dr]->on_path))
                       ALOOP(dr)
                       else
                       {
                          if (H_E[dr]->visited == WEAK)
                          {
                              H_E[dr]->visited = STRONG;
                              atpg_step(dr, level+1, UNDEFINED);
                          }
                       }
                   }
                 }
  
                /* SOURCE carrier */
                sr = H_S[j]->T.source->signal;
                if (H_S[sr]->attr != PI)
                {
                   if (! H_E[sr]->visited) atpg_step(sr, level+1, dr);
                   else
                   {
                      if ((H_E[sr]->comp == pres_comp) &&
                          (H_E[sr]->atpg_lev <= level) && (H_E[sr]->on_path))
                          ALOOP(sr)
                      else
                      {
                         if (H_E[sr]->visited == WEAK)
                         {
                            H_E[sr]->visited = STRONG;
                            atpg_step(sr, level+1, dr);
                         }
                      }
                   }
                 }
            }
            break;
      }
      pf = pf->c_next;
   }

   /* scan the connectivity list */
   while (pc != NULL)
   {
      /* adjacent carrier */
      if (H_S[pc->signal]->T.source->signal == el) 
                ad = H_S[pc->signal]->T.drain->signal;
      else
                ad = H_S[pc->signal]->T.source->signal;
      if ((ad != adj_n) && (H_S[ad]->attr != PI))
      {
         if (! H_E[ad]->visited)
         {
            H_E[ad]->visited = WEAK;
            atpg_step(ad, level, el);
         }
         else
         {
            if ((H_E[ad]->comp == pres_comp) && (H_E[ad]->atpg_lev < level)
                                             && (H_E[ad]->on_path)) ALOOP(ad)
            else
            {
               if (H_E[ad]->visited == WEAK)
               {
                   H_E[ad]->visited = STRONG;
                   atpg_step(ad, level, el);
               }
            }
         }
      }
      pc = pc->c_next;
   }
   /* pop from stack */
   H_E[el]->on_path = FALSE;
   if (atpg_flag) pop();
}


/*************************************************************************

   ROUTINE:     atpg_levelizing

   FUNCTION:    Assign a level for ATPG to each topological element

   AUTHOR(s):   Daniela Magnani

   DATE:        15/04/1987

   NOTES: 

   METHOD:      scan the table of the topological elements and call the
                function atpg_set_lev for the elements not_visited
                        a) Primary Inputs
			b) Flip-Flop

**********************************************************************/

atpg_levelizing()

{

/* local variables */

        unsigned long   i;

/* code */

        for (i = 0; i < h_exe.n_top; i++)
        {
                H_E[i]->visited = FALSE;
                H_E[i]->atpg_lev = 0; 
        }


        for (i = 0; i < h_exe.n_top; i++)
        {
                if (! H_E[i]->visited &&
		    (H_S[i]->attr == PI || 
		     (H_S[i]->code == GAT && (H_S[i]->G.type >= FIRST_FF &&
			                      H_S[i]->G.type <= LAST_FF))))
		{
                        atpg_set_lev(i, 0);
                }
        }
}


/*************************************************************************

   ROUTINE:     atpg_set_lev 

   FUNCTION:    Set atpg level for the element given as input.
                Scan fanout and connectivity lists assigning level 

   AUTHOR(s):   Daniela Magnani

   DATE:        21/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

atpg_set_lev(el, level)
unsigned long el;
unsigned long level;

{

/* local variables */

        CONN_PTR        pf;
        CONN_PTR        pc;
        unsigned long   j;
        unsigned long   dr;
        unsigned long   sr;
        unsigned long   ad;

/* code */
   
   H_E[el]->atpg_lev = level;
   H_E[el]->visited = TRUE;

   h_exe.max_atpg_lev = MAX(level, h_exe.max_atpg_lev);

   pc = NULL;
   switch (H_S[el]->code)
   {
      case GAT :
        pf = H_S[el]->G.l_fanout;
        break;
      case NOD :
      case WIR :
      case REG :
        pf = H_S[el]->C.l_fanout;
        pc = H_S[el]->C.l_conn;
        break;
      case CONST :
	switch (H_S[el]->I.type)
	{
	   case C_WIRE :
	        pf = H_S[el]->IW.l_fanout;
	        pc = H_S[el]->IW.l_conn;
 	        break;
	   case C_INT :
	   case D_CONST :
 		pf = H_S[el]->ID.l_fanout;
 	        break;
	   case C_CHAR :
 		pf = H_S[el]->IC.l_fanout;
 	        break;
	}
	break;
      case TRAN :
        if (H_S[el]->T.type != _DEPLETION)
        {     
	       pf = pc = NULL;
               /* DRAIN carrier */
               dr = H_S[el]->T.drain->signal;
               if (H_S[dr]->attr != PI && level >= H_E[dr]->atpg_lev)
			atpg_set_lev(dr, level+1);
               /* SOURCE carrier */
               sr = H_S[el]->T.source->signal;
               if (H_S[sr]->attr != PI && level >= H_E[sr]->atpg_lev)
			atpg_set_lev(sr, level+1);
        }
        break;
   }

   /* scan the fanout list */
   while (pf != NULL)
   {
      switch (H_S[j = pf->signal]->code)
      {
         case GAT :
		if (H_S[j]->G.type < FIRST_FF || H_S[j]->G.type > LAST_FF)
                {
            	    if (level >= H_E[j]->atpg_lev) atpg_set_lev(j, level+1);
		}
                break;
         case NOD :
         case WIR :
         case REG :
            if (level >= H_E[j]->atpg_lev) atpg_set_lev(j, level+1);
            break;
         case TRAN :
            if (H_S[j]->T.type != _DEPLETION)
            {     
               if (level >= H_E[j]->atpg_lev) atpg_set_lev(j, level+1);
	    }
	    break;
      }
      pf = pf->c_next;
   }
 
   /* scan the connectivity list */
   while (pc != NULL)
   {
      /* examine the adjacent carrier */
      if (H_S[pc->signal]->T.source->signal == el) 
                ad = H_S[pc->signal]->T.drain->signal;
      else
                ad = H_S[pc->signal]->T.source->signal;
      if ((H_S[ad]->attr != PI) && (level > H_E[ad]->atpg_lev))
		 atpg_set_lev(ad, level);
      pc = pc->c_next;
   }
}


/*************************************************************************

   ROUTINE:     cfunc

   FUNCTION:    create modules and functions C tables.

   AUTHOR(s):   Daniela Magnani

   DATE:        03/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

cfunc(el, list)
unsigned long el;
CONN_PTR list;

{

/* external functions */

	HFC_PTR		in_func();
	HMC_PTR		in_modc();
	char		*strsave();

/* local variables */

	FUNC_PTR	f_p;
	MODC_PTR	m_p;
	char		modf[NAMELENGTH + 1];
	char		*modfname;
	unsigned char	flag;
	HFC_PTR		f;
	HMC_PTR		m;

/* code */

	while (list != NULL)
	{
		if (list->type == CALL)
		{
			if (!func_flag)
			{
				int		i;
			 	for (i = 0; i < DIM_MODCHASH; i++)
					mod_headhash[i] = NULL; 
			 	for (i = 0; i < DIM_FUNCHASH; i++)
					fun_headhash[i] = NULL; 
				func_flag = TRUE;
			}

			f_p = list->CL.fun_p;
			m_p = f_p->p_mod;
			sprintf(modf, "%s%d", m_p->name, f_p->func);
			modfname = strsave(modf);
			flag = FALSE;
			f = in_func(modfname, &flag, f_p);
			if (flag)
			{
				m = in_modc(m_p->name, f->n_par, f->func);
				f->mod = m->n_mod;
			}
			list->signal = f->n_fun;
			list->CC.first = 0;
			list->CC.second = f->n_par;
		}
		list = list->c_next;
	}
}


/*************************************************************************

   ROUTINE:     in_func

   FUNCTION:    insert a C function description into the corresponding hash
		table

   AUTHOR(s):   Daniela Magnani

   DATE:        06/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

HFC_PTR in_func(s, f, f_p)
char *s;
char *f;
FUNC_PTR f_p;

{

/* external functions */

	long	hash();
	char	*alloca();
	long	strcmp();

/* local variables */

	long	i;
	HFC_PTR	p;

/* code */

        for (p = fun_headhash[i = hash(s) % DIM_FUNCHASH];
                p != NULL; p = p->next)
        {
                if (strcmp(s, p->name) == 0) return(p);
        }
        /* function not found  - new function creation */
        p = (HFC_PTR) alloca(sizeof(*p));

	*f = TRUE;
	p->name = s;	
        p->func = f_p->func;
        p->n_par = f_p->n_par;
	p->res_dim = f_p->res_dim;
        p->n_fun = h_exe.n_func++;
	p->insert = NULL;
	
	/* link FIFO to insertion list */
	if (last_f == NULL) 
		first_f = last_f = p;
	else
		last_f = last_f->insert = p;

        /* link LIFO to hash functions table */
        p->next = fun_headhash[i];
        fun_headhash[i] = p;
        return(p);
}


/*************************************************************************

   ROUTINE:     in_modc

   FUNCTION:    insert a C module description into the corresponding hash
		table

   AUTHOR(s):   Daniela Magnani

   DATE:        06/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

HMC_PTR in_modc(s,  n_par, n_fun)
char *s;
unsigned long n_par;
unsigned long n_fun;

{

/* external functions */

	long	hash();
	char	*alloca();
	long	strcmp();

/* local variables */

	long	i;
	HMC_PTR	p;

/* code */

        for (p = mod_headhash[i = hash(s) % DIM_MODCHASH];
                p != NULL; p = p->next)
        {
                if (strcmp(s, p->name) == 0)
		{
			p->n_func++;
			p->n_par = MAX(p->n_par, n_par);
			p->max_func = MAX(p->max_func, n_fun);
			return(p);
		}
        }
        /* module not found  - new module creation */
        p = (HMC_PTR) alloca(sizeof(*p));

	p->name = s;	
        p->n_func = 1;
        p->n_par = n_par;
        p->n_mod = h_exe.n_modc++;
	p->max_func = n_fun;
	p->insert = NULL;
	
	/* link FIFO to insertion list */
	if (last_m == NULL) 
		first_m = last_m = p;
	else
		last_m = last_m->insert = p;

        /* link LIFO to hash modules table */
        p->next = mod_headhash[i];
        mod_headhash[i] = p;
        return(p);
}


/*************************************************************************

   ROUTINE:     val_dc

   FUNCTION:    Control range of delay values (given as constant)
		Delay scaling.
		Returns the index of the constant element created

   AUTHOR(s):   Daniela Magnani

   DATE:        27/01/1989

   NOTES: 	t_un is the scale factor that must be applied
		val1 is the topological index of a constant

   METHOD: 

**********************************************************************/

long val_dc(val1, t_un)
unsigned long val1;
long t_un;
{

/* external functions */

	CONN_PTR	new_c_el();
	CONN_PTR	ins_last();
	char		*strsave();
	STAT_PTR	ret_ptr_exp_ex();
	long		uni_delay();

/* local variables */

	STAT_PTR	pc, ps;
	CONN_PTR	l_conn, l_conn1;
	char		found = FALSE;
	long		ret_v;
	long		e_val, e_val1;
	AUX_ST_PTR	buf_pt, aux_buf, aux_buf_old;
	char		as_val[255];
	unsigned long	n_el;

/* code */

    /* case 1 - No scaling required - only value checks are made
       case 2 - Scaling - new constant creation and value checks
       we look for an already created delay constant (of D_CONST type only)
       which implements the required integration */

     /* pc pointer to the input constant - ps pointer to the output constant*/
     /* val1 index of the input constant - ret_v index of the output one */
     /* e_val value of the output constant */

     ps = pc = ret_ptr_exp_ex(val1);
     if (pc->I.type != C_INT && pc->I.type != D_CONST)
	bug("Illegal constant type used to express delays");
     ret_v = val1;
     e_val = pc->ID.evval;

     if (t_un != 0)
     {
        /* case 2 */
	/* scan fanout of the constant val1 in order to find a D_CONST 
           constant which already implements the required operation */
	for (l_conn = pc->ID.l_fanout; l_conn != NULL && !found;
	     l_conn = l_conn->c_next)
	{
	   ps = ret_ptr_exp_ex(l_conn->signal);
	   if (ps->I.type == D_CONST)
	   {
		/* Element found if it accomplishes val1 * scaling factor */
		l_conn1 = ps->ID.l_fanin;
		if (l_conn1->c_next == NULL && l_conn1->signal == val1)
		{
			/* examine polish */
			l_conn1 = ps->ID.l_code;
			if (l_conn1->type == OPERAND && 
			    l_conn1->c_next->type == IMMEDIATE &&
			    l_conn1->c_next->c_next->type == OPERATOR &&
			    l_conn1->c_next->c_next->signal == M_MPY &&
			    l_conn1->c_next->signal == uni_delay(1, t_un))
			{
				found = TRUE;
				ret_v = l_conn->signal;
				e_val = ps->ID.evval;
			}
	        }
	   }
	}
	if (! found)
	{
		cre_exp_el(&ps, &n_el);
	
		ps->attr = PI;
		ps->code = CONST;
                ps->path_ptr = pc->path_ptr;

		ps->I.type = D_CONST;
		ps->ID.l_fanout = NULL;
		/* FANIN - format of the fanin list of the RT ass */
		ps->ID.l_fanin = new_c_el(pc->code, val1, 0, C_DATA);
		/* POLISH : operand - offset 0 in the fanin list
		   start bit 0 - stop bit constant length - 1 */
		ps->ID.l_code = new_c_el(OPERAND, 0, 0, pc->I.leng - 1);
		/* delay scaling */
		e_val1 = uni_delay(1, t_un);
		ps->ID.l_code = ins_last(ps->ID.l_code, 
			new_c_el(IMMEDIATE, e_val1, C_INT, 0));
                if (e_val1 > MAX_SHORTINT)
 		  ps->ID.l_code = ins_last(ps->ID.l_code, 
			 new_c_el(OPERATOR, _NOP, 0, 0));
             
		/* operator M_MPY */
		ps->ID.l_code = ins_last(ps->ID.l_code, 
			new_c_el(OPERATOR, M_MPY, 1, 0));
		/* STOP CODE generation */
		ps->ID.l_code = ins_last(ps->ID.l_code, 
			new_c_el(STOP_C, STOP_AS, 1, 0));

		/* update the fanout of the element val */
		l_conn = new_c_el(CONST, n_el, 0, 0);
		pc->ID.l_fanout = (pc->ID.l_fanout == NULL)
			? l_conn
			: ins_last(pc->ID.l_fanout, l_conn);

		/* value evaluation as e_val * e_val1 */
		if (e_val > (MAX_UNSIGNED / e_val1))
			e_val = MAX_UNSIGNED;
		else
			e_val *= e_val1;		

		ret_v = n_el;
	}
   }
   if (e_val < DELAY_MIN)
   {
   	e_val = DELAY_MIN;
        if (! del_ovf)
	{
		warn("delay underflow");
                h_exe.warn_cnt++;
		del_ovf = TRUE;
	}
    }
    if (e_val > DELAY_MAX)
    {
   	e_val = DELAY_MAX;
        if (! del_ovf)
	{
		warn("delay overflow");
                h_exe.warn_cnt++;
		del_ovf = TRUE;
	}
    }

    sprintf(as_val, "%d", e_val);
    ps->I.val = strsave(as_val);
    ps->I.leng = strlen(as_val);
    ps->ID.evval = e_val;
	
    h_exe.max_d = MAX(h_exe.max_d, e_val);
    return(ret_v);
}


/*************************************************************************

   ROUTINE:     ret_ptr_exp_ex

   FUNCTION:    Return the pointer to the topological descriptor
		for an element which may be in the expansion list
		or into the topological standard structure

   AUTHOR(s):   Daniela Magnani

   DATE:        19/09/1986

   NOTES: 	

   METHOD: 

**********************************************************************/

STAT_PTR ret_ptr_exp_ex(val1)
long val1;
{

/* local variables */

	AUX_ST_PTR	buf_pt;
	STAT_PTR	pc;

/* code */

	if (val1 >= h_exe.n_top)
	{
		/* scan the expression list in order to find the
		   element val1 */
		for (buf_pt = first; buf_pt->n_el != val1; 
			     		buf_pt = buf_pt->next);
		pc = buf_pt->st;
	}
	else
		pc = H_S[val1];

	return(pc);
}


/************************************************************************

   ROUTINE:     integ_static_exp

   FUNCTION:    Integrate the static elements contained into the
		expansion list into the static block

   AUTHOR(s):   Daniela Magnani

   DATE:        19/09/1986

   NOTES: 	

   METHOD: 

**********************************************************************/

integ_static_exp()

{

/* external functions */

        STAT_PTR        *new_top_el();
	STAT_EV_PTR	*new_ev_top_el();

/* local variables */

	int		i;
	AUX_ST_PTR	aux, aux_old;
	STAT_PTR 	*p_s;
	STAT_EV_PTR 	*p_ev_s;

/* code */

        /* creazione nuova struttura topologica 
	   free the old one and the expansion buffer */
	p_s = new_top_el(new_index, h_exe.n_top);
        p_ev_s = new_ev_top_el(new_index);
        for (i = 0; i < h_exe.n_top; i++)
	{
		p_s[i] = H_S[i];
		p_ev_s[i] = H_E[i];
	}
        for (aux = first, aux_old = NULL; aux != NULL; 
	     aux_old = aux, aux = aux->next)
	{
		if (aux_old != NULL) free(aux_old);
        	p_s[aux->n_el] = aux->st;
                p_ev_s[aux->n_el] = aux->st_ev;
	}
	if (aux_old != NULL) free(aux_old);
	h_exe.n_top = new_index;

	free(H_S);
        H_S = p_s;
	free(H_E);
        H_E = p_ev_s;
}


/************************************************************************

   ROUTINE:     cre_exp_el

   FUNCTION:    Create a static element into the expansion block.
		
   AUTHOR(s):   Daniela Magnani

   DATE:        19/09/1986

   NOTES: 	This element may be:
		AS_TR
		AS_TR_CN
		BIT_EX
		CONST
		We set the delay pointer to (0,0)
   METHOD: 

**********************************************************************/

cre_exp_el(p_s, n_el)
STAT_PTR *p_s;
long *n_el;

{

/* external functions */

        char            *alloca();
        STAT_PTR        new_top_des();
	STAT_EV_PTR	new_ev_top_des();

/* local variables */
        
        AUX_ST_PTR      buf_pt,
			aux_buf,
			aux_buf_old;

/* code */

	buf_pt = (AUX_ST_PTR) alloca(sizeof(*buf_pt));
	*n_el = buf_pt->n_el = new_index++;
	if (new_index > MAX_TOP) fatal("Too many topological elements");
	*p_s = buf_pt->st = new_top_des();
        buf_pt->st_ev = new_ev_top_des();
        buf_pt->st_ev->pdelay = p_0_d;
	buf_pt->next = NULL;

	/* update (FIFO) the exp list */
	if (first == NULL) 
		first = buf_pt;
	else
	{
		for (aux_buf = first; aux_buf!= NULL;
		     aux_buf_old = aux_buf, aux_buf = aux_buf->next);
			aux_buf_old->next = buf_pt;
	}
}

