/* LANGUAGE:    C        VAX VMS COMPILER VERSION 2.0-003


 PACKAGE NAME:          LGEN

 PACKAGE ROUTINES:      output_step()
                        gen_header()
                        gen_delay()
                        gen_static()
                        gen_conn()
                        gen_symtab()
                        gen_init_tab()
			gen_c_tab()
			gen_net()
			pr_tab()
			check_strength()

 PACKAGE FUNCTION:      create the linker output file .DEE
			and the netlist .EXP

 PACKAGE AUTHOR(s):     Daniela Magnani

 PACKAGE CREATION DATE: 20/08/1986

**********************************************************************/

# include <stdio.h>
# include "ldef.h"
# include "lstruct.h"

/***************************************************************************/
/*                                                                         */
/*                          VARIABLES DECLARATION                          */
/*                                                                         */
/***************************************************************************/

extern EHEADER          h_exe;

static FILE             *f;
static unsigned long    bit1d;
static char		fexe[NAMELENGTH + 5];
static long		head_symt_addr;	/* symbol table address */
static long		head_modt_addr;	/* module table address */


/*************************************************************************

   ROUTINE:     output_step

   FUNCTION:    generation step main

   AUTHOR(s):   Daniela Magnani

   DATE:        21/09/1988

   NOTES: 

   METHOD: 

**********************************************************************/

output_step()
{

/* external functions */

        FILE    *fopen();
        int     fclose();

/* external variables */

	extern	HDELAY_PTR	first_d;

/* local variables */

	HDELAY_PTR	aux;

/* code */
        
        sprintf(fexe, "%s%s", h_exe.name, TOP_EXE);
        
        /* open file .DEE */
        if ((f = fopen(fexe, "w")) == NULL)bug2("linker can't open file", fexe);

        gen_header();
        gen_delay();
	if (h_exe.n_modc > 0) gen_c_tab();
        gen_static();

        if (h_exe.n_init > 0) gen_init_tab();

	/* free the delay structure */
/*	while (first_d != NULL)
	{
		aux = first_d;
		first_d = first_d->insert;
		free(aux);
	}
*/

        gen_symtab();

        if (fclose(f) != 0) bug2("linker can't close file", fexe);

	if (h_exe.f_drc)
	{
		sprintf(fexe, "%s%s", h_exe.name, NET_EXE);
		/* open file for net list */
		if ((f = fopen(fexe, "w")) == NULL)
			bug2("linker can't open file", fexe);
		gen_net();
        	if (fclose(f) != 0) bug2("linker can't close file", fexe);
	}
}


/*************************************************************************

   ROUTINE:     gen_header

   FUNCTION:    Generate the header section of the file .DEE

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

# define avg(a,b) ((b) ? (long)((float)a / (float)b * 1000) : 0)

gen_header()

{
/* external functions */

	long	ftell();

/* local variables */

        unsigned long   i;

/* code */

        fprintf(f, "%s\n", C_LREL);
        fprintf(f, "%s\n", h_exe.name);

        if ((head_symt_addr = ftell(f)) == UNDEFINED)
			bug2("Error generating file ", fexe);
        fprintf(f, "%8d\n", 0);

        if ((head_modt_addr = ftell(f)) == UNDEFINED)
			bug2("Error generating file ", fexe);
        fprintf(f, "%8d\n", 0);

        fprintf(f, "%d\n", h_exe.warn_cnt);
        fprintf(f, "%d %d\n", h_exe.n_pi, h_exe.n_bitpi);
        fprintf(f, "%d %d\n", h_exe.n_po, h_exe.n_bitpo);
        fprintf(f, "%d %d\n", h_exe.n_pb, h_exe.n_bitpb);
        fprintf(f, "%d\n",  h_exe.n_const);
        fprintf(f, "%d\n",  h_exe.n_node);
        fprintf(f, "%d %d\n", h_exe.n_wire, h_exe.n_bitwire);
        fprintf(f, "%d %d\n", h_exe.n_reg, h_exe.n_bitreg);
        fprintf(f, "%d\n", h_exe.n_tran);
        fprintf(f, "%d\n",  h_exe.n_gate);
        fprintf(f, "%d\n", h_exe.n_rtas);
        fprintf(f, "%d\n", h_exe.n_rtcn);
        fprintf(f, "%d\n", h_exe.n_mem);
        fprintf(f, "%d\n", h_exe.n_locvar);
        fprintf(f, "%d\n", h_exe.n_proc);
        fprintf(f, "%d\n", 0); /* for future use */
        fprintf(f, "%d\n", h_exe.n_gtar);
        fprintf(f, "%d %d %d %d\n", h_exe.n_delay, h_exe.time_unit,
					h_exe.min_d, h_exe.max_d);
        fprintf(f, "%d\n", h_exe.n_init);
	fprintf(f, "%d %d\n", 0, 0); /* for future use */
	fprintf(f, "%d %d\n", h_exe.n_modc, h_exe.n_func);
        fprintf(f, "%d %d\n", h_exe.max_strength, h_exe.max_size);
        fprintf(f, "%d %d\n", h_exe.max_fan, h_exe.max_atpg_lev);
        fprintf(f, "%d %d\n", h_exe.vdd, h_exe.gnd);
        fprintf(f, "%d\n", h_exe.l_date);
        fprintf(f, "%d\n", h_exe.l_time);
        fprintf(f, "%d\n", h_exe.l_ctime);
        fprintf(f, "%d\n", h_exe.sb_dim);
        fprintf(f, "%d %d %d %d\n", h_exe.f_atpg, h_exe.f_sim, h_exe.f_drc,
				    h_exe.f_case);
        fprintf(f, "%d %d %d %d\n", h_exe.n_top,
                                h_exe.n_fanin,
                                h_exe.n_fanout,
                                h_exe.n_conn);
        i = h_exe.n_node;
        fprintf(f, "%d %d %d\n", avg(h_exe.n_nfanout, i),
                                avg(h_exe.n_nfanin, i),
                                avg(h_exe.n_nconn, i));
        i = h_exe.n_wire;
        fprintf(f, "%d %d\n", avg(h_exe.n_wfanout, i),
                                avg(h_exe.n_wfanin, i));
        i = h_exe.n_reg;
        fprintf(f, "%d %d\n", avg(h_exe.n_rfanout, i),
                                avg(h_exe.n_rfanin, i));        
        i = h_exe.n_gate;
        fprintf(f, "%d %d\n", avg(h_exe.n_gfanout, i),
                                avg(h_exe.n_gfanin, i));
        i = h_exe.n_rtas;
        fprintf(f, "%d %d\n", avg(h_exe.n_asfanout, i),
                                avg(h_exe.n_asfanin, i));       
        i = h_exe.n_rtcn;
        fprintf(f, "%d %d\n", avg(h_exe.n_cnfanout, i),
                                avg(h_exe.n_cnfanin, i));       
}


/*************************************************************************

   ROUTINE:     gen_delay

   FUNCTION:    Generate the delay section of the file .DEE
		scanning the delay list

   AUTHOR(s):   Daniela Magnani

   DATE:        01/06/1988

   NOTES: 

   METHOD: 

**********************************************************************/

gen_delay()

{
/* external variables */

        extern	HDELAY_PTR       first_d;

/* local variables */

	HDELAY_PTR	aux;
	unsigned long	rise;
	unsigned long	fall;

/* code */
        
	aux = first_d;
	while (aux != NULL)
	{
		rise = aux->rise;
		fall = aux->fall;
		if (aux->rise_type == DEL_CON)
		   rise |= 0x80000000;
		if (aux->fall_type == DEL_CON)
		   fall |= 0x80000000;
		fprintf(f, "%d %d\n", rise, fall);
		aux = aux->insert;
        }
}


/*************************************************************************

   ROUTINE:     gen_static

   FUNCTION:    Generate the topological section of the file .DEE

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 	

   METHOD: 

**********************************************************************/

gen_static()

{

/* external functions */

	char	check_strength();

/* local variables */

        unsigned long   i;
        unsigned long   fi;
        unsigned long   fi_t;
        unsigned long   fo;
        unsigned long   co;
        unsigned long   cd;
        unsigned long   delay;
	unsigned char	size;

/* code */

        for (i = 0; i < h_exe.n_top; i++)
        {
                fprintf(f, "%d %d %d %d\n",
                        H_S[i]->attr,
                        H_S[i]->code,
                        H_E[i]->sim_lev,
                        H_E[i]->atpg_lev);
                switch (H_S[i]->code)
                {
                        case NOD :
                                delay = (H_S[i]->attr == PI) 
                                        ? 0 : H_E[i]->pdelay->n_delay;
                                fprintf(f, "%d %d %d %d %d %d\n",
                                        H_S[i]->C.dim,
                                        H_S[i]->C.size,
                                        delay,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        co = H_E[i]->n_conn);
                                if (fo > 0) gen_conn(H_S[i]->C.l_fanout, NULL);
                                if (fi > 0) gen_conn(H_S[i]->C.l_fanin, NULL);
                                if (co > 0) gen_conn(H_S[i]->C.l_conn, NULL);
                                break;
			case WIR :
			case REG :
				size = (check_strength(H_S[i]->C.l_fanin))
					? H_S[i]->C.size + 128 : H_S[i]->C.size;
                                delay = (H_S[i]->attr == PI) 
                                        ? 0 : H_E[i]->pdelay->n_delay;
                                fprintf(f, "%d %d %d %d %d %d\n",
                                        H_S[i]->C.dim,
                                        size,
                                        delay,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        co = H_E[i]->n_conn);
                                if (fo > 0) gen_conn(H_S[i]->C.l_fanout, NULL);
                                if (fi > 0) gen_conn(H_S[i]->C.l_fanin, NULL);
                                if (co > 0) gen_conn(H_S[i]->C.l_conn, NULL);
                                break;
			case LOCV :
                                fprintf(f, "%d %d %d %d\n",
                                        H_S[i]->C.del_type,
                                        H_S[i]->C.dim,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin);
                                if (fo > 0) gen_conn(H_S[i]->C.l_fanout, NULL);
                                if (fi > 0) gen_conn(H_S[i]->C.l_fanin, NULL);
                                break;
                        case GAT :
                                fprintf(f, "%d %d %d %d %d %d\n",
                                        H_S[i]->G.type,
                                        H_E[i]->pdelay->n_delay,
                                        H_S[i]->G.str_0,
                                        H_S[i]->G.str_1,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin);
                                if (fo > 0) gen_conn(H_S[i]->G.l_fanout, NULL);
                                if (fi > 0) gen_conn(H_S[i]->G.l_fanin, NULL);
                                break;
                        case TRAN :
                                fprintf(f, "%d %d\n",
                                        H_S[i]->T.type,
                                        H_S[i]->T.str);
                                gen_conn(H_S[i]->T.drain, NULL);
                                gen_conn(H_S[i]->T.gate, NULL);
                                gen_conn(H_S[i]->T.source, NULL);
                                break;
                        case RT_AS :
                        case RT_LD :
                                fprintf(f, "%d %d %d %d %d %d %d\n",
                                        bit1d = H_S[i]->A.bit,
                                        H_E[i]->pdelay->n_delay,
                                        H_S[i]->A.str_0,
                                        H_S[i]->A.str_1,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        cd = H_E[i]->n_code);
                                if (fo > 0) gen_conn(H_S[i]->A.l_fanout, FANOUT);
                                if (fi > 0) gen_conn(H_S[i]->A.l_fanin, NULL);
                                if (cd > 0) gen_conn(H_S[i]->A.l_code, CODE);
                                break;
                        case BIT_EX :
                                fprintf(f, "%d %d %d %d %d %d %d\n",
                                        bit1d = H_S[i]->A.bit,
                                        H_E[i]->pdelay->n_delay,
                                        H_S[i]->A.str_0,
                                        H_S[i]->A.str_1,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        cd = H_E[i]->n_code);
                                if (fo > 0) gen_conn(H_S[i]->A.l_fanout, BIT_EX);
                                if (fi > 0) gen_conn(H_S[i]->A.l_fanin, NULL);
                                if (cd > 0) gen_conn(H_S[i]->A.l_code, CODE);
                                break;
                        case AS_TR :
                                fprintf(f, "%d %d %d %d %d %d %d\n",
                                        bit1d = H_S[i]->A.bit,
                                        H_E[i]->pdelay->n_delay,
                                        H_S[i]->A.str_0,
                                        H_S[i]->A.str_1,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        cd = H_E[i]->n_code);
                                if (fo > 0) gen_conn(H_S[i]->A.l_fanout, NULL);
                                if (fi > 0) gen_conn(H_S[i]->A.l_fanin, NULL);
                                if (cd > 0) gen_conn(H_S[i]->A.l_code, CODE);
                                break;
                        case AS_TR_CN :
                                fprintf(f, "%d %d %d %d %d %d %d\n",
                                        bit1d = H_S[i]->A.bit,
                                        H_E[i]->pdelay->n_delay,
                                        H_S[i]->A.str_0,
                                        H_S[i]->A.str_1,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        H_E[i]->n_code);
                                if (fo > 0) gen_conn(H_S[i]->A.l_fanout, NULL);
                                if (fi > 0) gen_conn(H_S[i]->A.l_fanin, NULL);
				break;
                        case RT_CN :
                        case RT_CNT :
                                fprintf(f, "%d %d %d\n",
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        cd = H_E[i]->n_code);
                                if (fo > 0) gen_conn(H_S[i]->O.l_fanout, NULL);
                                if (fi > 0) gen_conn(H_S[i]->O.l_fanin, NULL);
                                if (cd > 0) gen_conn(H_S[i]->O.l_code, CODE);
                                break;
                        case CONST :
				if (H_S[i]->I.type == D_CONST)
				    H_S[i]->I.type = C_INT;
                                fprintf(f, "%d %d %d %d %d %s\n",
					H_S[i]->I.type,
                                        fo = H_E[i]->n_fanout,
                                        fi = H_E[i]->n_fanin,
                                        co = H_E[i]->n_conn,
                                        cd = H_E[i]->n_code,
                                        H_S[i]->I.val);
				if (H_S[i]->I.type == C_WIRE)
				{
                                  if (fo > 0)
					gen_conn(H_S[i]->IW.l_fanout, NULL);
	                          if (co > 0)
					gen_conn(H_S[i]->IW.l_conn, NULL);
				}
				if (H_S[i]->I.type == C_INT)
				{
                                  if (fo > 0)
					gen_conn(H_S[i]->ID.l_fanout, NULL);
                                  if (fi > 0)
					gen_conn(H_S[i]->ID.l_fanin, NULL);
                                  if (cd > 0)
					gen_conn(H_S[i]->ID.l_code, NULL);
				}
				if (H_S[i]->I.type == C_CHAR)
				{
                                  if (fo > 0)
					gen_conn(H_S[i]->IC.l_fanout, NULL);
				}
                                break;
                        case MEM :
                                fprintf(f, "%d %d %d %d %d %d %d %s\n",
                                        H_S[i]->M.type,
                                        bit1d = H_S[i]->M.word_size,
                                        H_S[i]->M.addr_size,
                                        H_S[i]->M.str_0,
                                        H_S[i]->M.str_1,
                                        H_E[i]->pdelay->n_delay,
                                        fi = H_E[i]->n_fanin,
                                        H_S[i]->M.val);
                                gen_conn(H_S[i]->M.l_fanout, FANOUT);
                                if (fi > 0) gen_conn(H_S[i]->M.l_fanin, NULL);
                                break;
			case PROC :
                                fprintf(f, "%d %d %d %d %d %d\n",
                                        H_S[i]->P.type,
                                        H_E[i]->n_fanout,
                                        fi_t = H_E[i]->n_tfanin,
                                        fi = H_E[i]->n_fanin,
                                        H_E[i]->n_conn,
                                        cd = H_E[i]->n_code);
                                  if (fi_t > 0)
					gen_conn(H_S[i]->P.l_tfanin, NULL);
                                  if (fi > 0)
					gen_conn(H_S[i]->P.l_dfanin, NULL);
                                  if (cd > 0)
					gen_conn(H_S[i]->P.l_code, CODE);
				break;
                }
        }
}


/*************************************************************************

   ROUTINE:     gen_conn

   FUNCTION:    generate the variable part of a topological descriptor

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

gen_conn(p_conn, list_type)
CONN_PTR p_conn;
long list_type; /* FANOUT or CODE or NULL */

{

/* local variables */

        unsigned long   bit2d;

/* code */
        
        while (p_conn != NULL)
        {
		if (list_type == CODE && p_conn->type == TIME)
		{
			p_conn->signal = p_conn->CT.del_p.delh_ptr->n_delay;
			p_conn->CC.first = 0;
			p_conn->CC.second = 0;
		}
                fprintf(f, "%d %d", p_conn->type, p_conn->signal);
                switch (list_type)
                {
                  case FANOUT:
			bit2d = --bit1d;
                        bit1d -= (p_conn->CC.second - p_conn->CC.first);
                        fprintf(f, " %d %d %d %d\n",
                        		p_conn->CC.first,
                                        p_conn->CC.second,
                                        bit1d,
                                        bit2d);
                        break;
                  case BIT_EX:
                        fprintf(f, " %d %d 0 0\n",
                        		p_conn->CC.first,
                                        p_conn->CC.second);
                        break;
                  default:
			fprintf(f, " %d %d\n",
                                        p_conn->CC.first,
                                        p_conn->CC.second);
                        break;
                }
                p_conn = p_conn->c_next;
        }
}

/*************************************************************************

   ROUTINE:     gen_symtab()

   FUNCTION:    write the symbol table

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

gen_symtab()

{
/* external variables */

        extern UNIT_PTR         p_level[];
        extern long             max_level;
        extern IN_T_PTR         in_tree;

/* external functions */

        char            *alloca();
        long		ftell();
        int             fseek();
	
/* local variables */

        UNIT_PTR        u;
        INTF_I_PTR	p_i;
        INTF_PTR	p_m;
        INTF_PTR	p_aux;
        INTF_PTR        mod_tab;        /* start pointer to module table */
        INTF_PTR        *v;
        SYMT_PTR        p_s;
        long            i;
        long            j;
        long            i_d;
        long            mod_ind;        /* module table index */
        long		mod_tb_addr;    /* module table start address */
        long		sym_tb_addr;    /* global symbol table start address */
        long		mod_addr;       /* symbol table start address */

/* code */
        
	/* store the symbol table start address */
        if ((sym_tb_addr = ftell(f)) == UNDEFINED) 
			bug2("Error generating file ", fexe);

        /* write header record */
        fprintf(f, "%d %d\n", h_exe.n_inst, h_exe.n_def);
        
        /* write instantiation tree */
        fprintf(f, "%s %d %d %s %d %d %d %d\n",
        	in_tree[0].in_name,
                in_tree[0].st_addr,
                in_tree[0].end_addr,
                in_tree[0].def_name,
                in_tree[0].father,
                in_tree[0].child,
                in_tree[0].brother,
                in_tree[0].intf_dim);
        /* fprintf(f, "%s %d %s %d %d %d %d\n",
        	in_tree[0].in_name,
                in_tree[0].st_addr,
                in_tree[0].def_name,
                in_tree[0].father,
                in_tree[0].child,
                in_tree[0].brother,
                in_tree[0].intf_dim); */
        for (i = 1; i < h_exe.n_inst; i++)
        {        
	        fprintf(f, "%s %d %d %s %d %d %d %d\n",
        		in_tree[i].in_name,
	                in_tree[i].st_addr,
	                in_tree[i].end_addr,
        	        in_tree[i].def_name,
                	in_tree[i].father,
	                in_tree[i].child,
        	        in_tree[i].brother,
                	i_d = in_tree[i].intf_dim);
	      /*  fprintf(f, "%s %d %s %d %d %d %d\n",
        		in_tree[i].in_name,
	                in_tree[i].st_addr,
        	        in_tree[i].def_name,
                	in_tree[i].father,
	                in_tree[i].child,
        	        in_tree[i].brother,
                	i_d = in_tree[i].intf_dim);  */
                /* interface table */
                p_i = in_tree[i].intf_table;
		if (in_tree[i].un_addr->intf_sort_table == NULL)
		{
			/* model interface table has to be sorted.*/
			p_m = in_tree[i].un_addr->intf_table;
	        	if (i_d > 1)
			{
			   v = (INTF_PTR*) alloca(i_d * sizeof(INTF_PTR));
	        	   for (j = 0; j < i_d; j++)
		           {
        		        v[j] = p_m++;
		           }
			   quicksort(&v[0], &v[--j]);

			   p_aux = in_tree[i].un_addr->intf_sort_table = v[0];
	        	   for (j = 0; j < i_d - 1; j++)
		           {
        		       	v[j]->sort_link = v[j + 1];
		           }
			   v[i_d - 1]->sort_link = NULL;
			   free(v);
			}
			else
			{
			   in_tree[i].un_addr->intf_sort_table = p_m;
			   p_m->sort_link = NULL;
			}
		}

		/* We have to scan the sort link in order to write
		   the table sorted */
		p_aux = in_tree[i].un_addr->intf_table;
		for (p_m = in_tree[i].un_addr->intf_sort_table;
		     p_m != NULL; p_m = p_m->sort_link)
                {
                        fprintf(f, "%d %s\n", 
			(INTF_I_PTR)(p_i + (p_m - p_aux))->addr, 
			p_m->name);
                }
        }
        
        /* creation of the module table */
        p_m = mod_tab = (INTF_PTR) alloca(h_exe.n_def * sizeof(*mod_tab));
        mod_ind = 0;
        
        /* symbol tables (accessed via units definitions tree) */
        for (i = max_level; i >= 0 ; i--)
        {
                for (u = p_level[i]; u != NULL; u = u->level_next)
                {
                        /* write module name and file address into
                           the module table */
                        mod_tab[mod_ind].name = u->name;
			mod_tab[mod_ind].par_n = U_H->unit_type;
                        if ((mod_addr = ftell(f)) == UNDEFINED)
				bug2("Error generating file ",fexe);
                        mod_tab[mod_ind++].addr = mod_addr;
                        
                        /* symbol table */
                        fprintf(f, "%s %d\n", u->name, U_H->n_sym);
                        for (j = 0; j < U_H->n_sym; j++)
                        {
                                fprintf(f, "%d %d %s\n",
                                        U_T[j].st_el,
					U_T[j].el_type,
                                        U_T[j].el_name);
                        }
                }
        }
        
        /* sort module table */
        v = (INTF_PTR*) alloca(h_exe.n_def * sizeof(INTF_PTR));
        for (j = 0; j < h_exe.n_def; j++)
        {
                v[j] = p_m++;
        }
        if (h_exe.n_def > 1) quicksort(&v[0], &v[--j]);
        
        /* write module table */
        if ((mod_tb_addr = ftell(f)) == UNDEFINED) 
			bug2("Error generating file ", fexe);
        for (i = 0; i < h_exe.n_def; i++)
        {
                fprintf(f, "%s %d %d\n", v[i]->name, v[i]->par_n, v[i]->addr);
        }
        

        /* write in the header the symbol and module table address */
        if (fseek(f, head_symt_addr, 0) == EOF)
		bug2("Error generating file ",fexe);
        fprintf(f, "%8d", sym_tb_addr);
        if (fseek(f, head_modt_addr, 0) == EOF)
		bug2("Error generating file ",fexe);
        fprintf(f, "%8d", mod_tb_addr);

}  /* end of gen_symtab() */


/*************************************************************************

   ROUTINE:     gen_init_tab()

   FUNCTION:    generate the initialization table

   AUTHOR(s):   Daniela Magnani

   DATE:        1/10/1986

   NOTES: 

   METHOD: 

**********************************************************************/

gen_init_tab()

{
/* external variables */

        extern	INIT_PTR         init_headhash[];

/* external functions */


/* local variables */

        INIT_PTR        p;
        CONN_PTR        l_f;
        long            i;

/* code */
        
        for (i = 0; i < DIM_INITHASH; i++)
        {
                for (p = init_headhash[i]; p != NULL; p = p->next)
                {
                        fprintf(f, "%s %d\n", p->val, p->n_fanout);
                        for (l_f = p->l_fanout; l_f != NULL; l_f = l_f->c_next)
                                fprintf(f, "%d\n", l_f->signal);
                }
        }
}


/*************************************************************************

   ROUTINE:     gen_net

   FUNCTION:    generate the net list for DE/DRC

   AUTHOR(s):   Daniela Magnani

   DATE:        04/05/1987

   NOTES: 

   METHOD: 

**********************************************************************/
gen_net()

{
/* external variables */

	extern	UNIT_PTR	p_level[];
	extern	long		max_level;
	extern	IN_T_PTR	in_tree;
	extern  IN_ST		par_des[];
	extern	unsigned char	vdd_ext;
	extern	unsigned char	gnd_ext;

/* external functions */

	char			*ret_path();
	char			*ret_in();
	QUEUE_PTR		dequeue();
	char			*to_up();

/* local variables */

	UNIT_PTR		u;
	INTF_I_PTR		p_i;
	INTF_PTR		p_m;
	char			path[PATHLENGTH + 1];
	long			i;
	long			cc;
	long			j;
	unsigned long		old_ind;
	unsigned short		old_par;
	char			old_path[PATHLENGTH + 3];
	QUEUE_PTR		t_p;
	unsigned long		f_n;
	unsigned long		dir;
	CONN_PTR		conn;
	long			ty;
	unsigned char		flag;

/* code */

   fprintf(f,"%s %s\n", DRC_KW, h_exe.name);

   /* table of PI/PO/PB global */
   u = p_level[max_level];
   cc = h_exe.n_pi + h_exe.n_po + h_exe.n_pb;
   if (h_exe.n_po > 0)
   {
	pr_tab(u->po_table, h_exe.n_po, &cc);
   }
   if (h_exe.n_pi > 0)
   {
	pr_tab(u->pi_table, h_exe.n_pi, &cc);
   }
   if (h_exe.n_pb > 0)
   {
	pr_tab(u->pb_table, h_exe.n_pb, &cc);
   }

   /* print the net list of units Gate Array scanning the 
   instantiation tree */
   enqueue(0);
   while ((t_p = dequeue()) != NULL)
   {
      /* enqueue the children for units not Gate Array */
      dir = in_tree[i = t_p->el].un_addr->dir;
      if (dir == MAX_UNSIGNED || par_des[dir].type == DIR ||
	  par_des[dir].type == DIRL)
      {
		f_n = in_tree[i].child;
		while (f_n != MAX_UNSIGNED)
		{
			enqueue(f_n);
			f_n = in_tree[f_n].brother;
		}
      }
      else
      if (par_des[dir].type == LIBD)
      {
	  fprintf(f, "\n%s  %s%s", to_up(in_tree[i].def_name),
				       DRC_NM, ret_in(i, path));
	  p_i = in_tree[i].intf_table;
	  if (p_i != NULL)
	  {
	  	p_m = in_tree[i].un_addr->intf_table;
		old_par = (short)MAX_UNSIGNED;
		old_ind = MAX_UNSIGNED;
		sprintf(old_path, "%s", NULL_NAME);
		for (j = 0; j < in_tree[i].intf_dim; j++)
		{
			if (p_m->par_n != old_par)
			{
				fprintf(f, "%s\n", old_path);
				old_par = p_m->par_n;
				old_ind = p_i->addr;
				if (p_i->f_conn)
					  sprintf(old_path,"%s",DRC_NM);
				else
				{
				  sprintf(old_path,"%s%s", DRC_NM,
					  ret_path(old_ind, path));
				  if (H_S[p_i->addr]->code == GAT 
				  && H_E[p_i->addr]->n_fanout > 0)
		                  {
					conn = H_S[p_i->addr]->G.l_fanout;
					flag = FALSE;
		                        while (conn != NULL && !flag)
                       			{
                               			ty = conn->type;
                               			if (ty == NOD || ty == WIR)
		                                {
		                                        flag = TRUE;
							sprintf(old_path,"%s%s",
							 DRC_NM,
							ret_path(conn->signal,
								 path));
                       			        }
		                                conn = conn->c_next;
                       			}
		                  }
				}
			}
			p_i++;
			p_m++;
		}
		fprintf(f, "%s %s\n", DRC_LN, old_path);
	  }
      }
      /* free the space occupied by the element dequeued */
      free(t_p);
   }

   /* supply list */
   fprintf(f, "\n");
   if (GND != MAX_UNSIGNED & gnd_ext) fprintf(f, "SUPPLY0  **gnd ;");
   if (VDD != MAX_UNSIGNED & vdd_ext) fprintf(f, "SUPPLY1  **vdd ;");

   /* end of file */
   fprintf(f, "\n.\n");
}


/*************************************************************************

   ROUTINE:     pr_tab

   FUNCTION:    

   AUTHOR(s):   Daniela Magnani

   DATE:        04/05/1987

   NOTES: 

   METHOD: 

**********************************************************************/

pr_tab(list, n_el, counter)
long *list;
unsigned long n_el;
long *counter;

{

/* external functions */

	char 	*ret_path();

/* local variables */

	char	path[PATHLENGTH + 1];

/* code */

	while (n_el -- > 0)
	{
		if ((*counter)-- > 1)
			fprintf(f, "%s%s\n", DRC_NM, ret_path(*list++, path));
		else
			fprintf(f, "%s  %s%s\n", DRC_LN, DRC_NM, 
					ret_path(*list++, path));
	}
}


/*************************************************************************

   ROUTINE:     gen_c_tab()

   FUNCTION:    generate the modules & functions C table

   AUTHOR(s):   Daniela Magnani

   DATE:        06/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

gen_c_tab()

{
/* external variables */

        extern	HMC_PTR         first_m;
	extern  HFC_PTR		first_f;

/* external functions */


/* code */
        
	while (first_m != NULL)
	{
	        fprintf(f, "%s %d %d %d\n", first_m->name, first_m->n_func,
			first_m->max_func, first_m->n_par);
		first_m = first_m->insert;
	}
	while (first_f != NULL)
	{
		fprintf(f, "%d %d %d %d\n", first_f->func, first_f->n_par,
			first_f->mod, first_f->res_dim);
		first_f = first_f->insert;
        }
}


/*************************************************************************

   ROUTINE:     check_strength()

   FUNCTION:    scan the fanin list of a wire and return TRUE if all the
		elements writing onto the wire have the same strength
		Returns also TRUE if the wire has not fanin.

   AUTHOR(s):   Daniela Magnani

   DATE:        06/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

char check_strength(list)
CONN_PTR list;
{

/* local variables */

	unsigned char	str_flg = TRUE;
	unsigned char	str_set_flg = FALSE;
	unsigned char	str_val;

/* code */

	while (list != NULL && str_flg)
	{
		switch (H_S[list->signal]->code)
		{
			case GAT :
				if (! str_set_flg)
				{
				   str_set_flg = TRUE;
				   str_val = H_S[list->signal]->G.str_0;
				}
				if (str_val != H_S[list->signal]->G.str_0 ||
				    str_val != H_S[list->signal]->G.str_1)
					str_flg = FALSE;
				break;
			case RT_AS :
			case RT_LD :
			case AS_TR :
			case AS_TR_CN :
			case BIT_EX :
				if (! str_set_flg)
				{
				   str_set_flg = TRUE;
				   str_val = H_S[list->signal]->A.str_0;
				}
				if (str_val != H_S[list->signal]->A.str_0 ||
				    str_val != H_S[list->signal]->A.str_1)
					str_flg = FALSE;
				break;
			case MEM :
				if (! str_set_flg)
				{
				   str_set_flg = TRUE;
				   str_val = H_S[list->signal]->M.str_0;
				}
				if (str_val != H_S[list->signal]->M.str_0 ||
				    str_val != H_S[list->signal]->M.str_1)
					str_flg = FALSE;
				break;
			default :
				/* do nothing */
				break;
		}
		list = list->c_next;
	}
	return(str_flg);
}
