/* LANGUAGE:    C        VAX VMS COMPILER VERSION 2.0-003


 PACKAGE NAME:          LINDEX

 PACKAGE ROUTINES:	open_lib()
			close_lib()
			search()
			free_lib()

 PACKAGE FUNCTION:      library management functions called by the mozart 
			linker

 PACKAGE AUTHOR(s):     Daniela Magnani

 PACKAGE CREATION DATE: 22/07/1987

**********************************************************************/

# include <stdio.h>
# include "ldef.h"
# include "lstruct.h"


/***************************************************************************/
/*                                                                         */
/*                          VARIABLES DECLARATION                          */
/*                                                                         */
/***************************************************************************/

static FILE    	*f_m;
static char     res_date[26];
static char     aux_t[18];
static long     n_page;
static long     n_line;


/*************************************************************************

   ROUTINE:     open_lib

   FUNCTION:    open a library and transfer data for the library header and 
		index into main memory.
		This functions returns TRUE is the open was successfull,
		FALSE otherwise.

   AUTHOR(s):   Daniela Magnani

   DATE:        29/07/1987

   NOTES:	

   METHOD: 

**********************************************************************/

char open_lib(s, l, mode)
char *s;
LIB_PTR l;
char *mode;

{
/* external functions */

	FILE	*fopen();
	int	fclose();
	int	fseek();
	int	fscanf();
	char	*alloca();
	int	strlen();

/* local variables */

	char		aux[FILELENGTH + 1];
	FILE		*f;
	IND_PTR		p;
	long		n_el;
	char		un_n[IDLENGTH + 1];

/* body */

	sprintf(aux, "%s%s", s, LIB_EXT);
	
	if ((f = fopen(aux, mode)) == NULL)
	{
		warn2("library manager can't open the library", s);
		return(FALSE);
	}

	/* library successfully opened - fill the library structure */
	l->name = (char*) alloca((strlen(s) + 1) * sizeof(char));

	sprintf(l->name, "%s", s);
	l->f = f;
	l->mod_flag = FALSE;

	/* load the library header */
	fscanf(f, "%d", &l->head_pos);
	if (fseek(f, l->head_pos, 0) == EOF) POS_ERR;	
	fscanf(f, "%ld %ld %ld %ld %ld %ld %ld %ld",
		&l->head.cell,
		&l->head.cell_d,
		&l->head.c_time,
		&l->head.a_time,
		&l->head.m_time,
		&l->head.rel_num,
		&l->head.data_type,
		&l->head.index_pos);

	l->pos_free = l->head.index_pos;

	/* load the index table */
	l->table.exp_list = NULL;
	l->table.table = NULL;
	n_el = l->table.array_el = l->head.cell;
	if (n_el > 0)
	{
		if (fseek(f, l->head.index_pos, 0) == EOF) POS_ERR;	
		l->table.table = p = (IND_PTR) alloca(n_el * sizeof(*p));
		while (n_el-- > 0)
		{
			fscanf(f, "%s %c %ld %ld %ld %ld %ld",
				p->in_d.key,
				&p->in_d.val_flag,
				&p->in_d.p_deo,
				&p->in_d.dim_deo,
				&p->in_d.c_time,
				&p->in_d.a_time,
				&p->in_d.m_time);
			p->next = NULL;
			p++;
		}
	}

	return(TRUE);
}


/*************************************************************************

   ROUTINE:     close_lib

   FUNCTION:    close a library.
		If the library contents has been modified the header and the
		index are rewritten onto the file.
		Otherwise only the last access time is modified.

   AUTHOR(s):   Daniela Magnani

   DATE:        29/07/1987

   NOTES:

   METHOD: 

**********************************************************************/

char close_lib(l)
LIB_PTR l;

{

/* external functions */

	long	ftell();
	int	fseek();
	int 	fclose();
	long	time();
	char	free_lib();

/* local variables */

	unsigned long 	i;
	IND_PTR		p;

/* body */

	time(&l->head.a_time);

	/* write header onto file */
	if (fseek(l->f, l->head_pos, 0) == EOF) POS_ERR;	
	fprintf(l->f, "%ld %ld %ld %ld %ld %ld %ld %ld\n",
		l->head.cell,
		l->head.cell_d,
		l->head.c_time,
		l->head.a_time,
		l->head.m_time,
		l->head.rel_num,
		l->head.data_type,
		l->head.index_pos);
	
	if (fclose(l->f) != 0)
	{
		warn2("library manager can't close the library", l->name);
		return(FALSE);
	}
	return(free_lib(l));
}


	
/*************************************************************************

   ROUTINE:     search

   FUNCTION:    search a unit into an index

   AUTHOR(s):   Daniela Magnani

   DATE:        03/08/1987

   NOTES: 

   METHOD: 

**********************************************************************/

IND_PTR search(s, l, p)
char *s;	/* name of the unit to search */
LIB_PTR l;
IND_PTR *p;
{

/* external functions */

	int		strcmp();

/* local variables */
	
	long		cond;
	IND_PTR		high;
	IND_PTR		low;
	IND_PTR		half;

/* body */

	*p = NULL;
	if (l->table.table != NULL)
	{
		if (strcmp(s, l->table.table->in_d.key) >= 0)
		{
			/* binary search */
			high = (IND_PTR)
				 ((l->table.table) + (l->table.array_el - 1));
			low = l->table.table;
			while (low <= high)
			{
				half = low + (high - low) / 2;
				if ((cond = strcmp(s, half->in_d.key)) < 0)
					high = half - 1;
				else if (cond > 0)
					low = half + 1;
				else
					return(half);
			}
		}
	}
	return(NULL);
}


/*************************************************************************

   ROUTINE:     free_lib

   FUNCTION:    free the space allocated for a library

   AUTHOR(s):   Daniela Magnani

   DATE:        23/10/1987

   NOTES: 

   METHOD: 

**********************************************************************/

char free_lib(l)
LIB_PTR l;

{

/* body */

	if (l->table.table != NULL) free(l->table.table);
	free(l->name);
	return(TRUE);
}

