/* LANGUAGE:    C        VAX VMS COMPILER VERSION 2.0-003


 PACKAGE NAME:          LLOAD

 PACKAGE ROUTINES:      load_unit()
                        load_header()
                        load_delay()
                        load_static()
                        load_conn()
                        load_symtab()
                        load_init()
			load_unitc()
                        new_top_el()
                        new_top_des()
			new_ev_top_el()
			new_ev_top_des()
			free_static()
			create_stat_block()

 PACKAGE FUNCTION:      read and load files .DEO

 PACKAGE AUTHOR(s):     Daniela Magnani

 PACKAGE CREATION DATE: 20/08/1986

**********************************************************************/

# include <stdio.h>
# include "ldef.h"
# include "lstruct.h"

/***************************************************************************/
/*                                                                         */
/*                          VARIABLES DECLARATION                          */
/*                                                                         */
/***************************************************************************/

static OHEADER_PTR      p_h;
static FILE             *f;
static long             *p_ld;
static ODELAY_PTR       delay_unit;	/* delay table */
static char             fobj[FILEDEF + 1];
static MODC_PTR		modc_unit;
static FUNC_PTR		func_unit;

/* static elements allocation */

STAT_PTR	free_stat_list = NULL;
long		free_stat_cnt = 0;
long		block_stat_cnt = 0;

/* header variables */

unsigned long		u_n_bitpi,
			u_n_bitpo,
			u_n_bitpb,
			u_n_const,
			u_n_tran,
			u_n_gate,
			u_n_rtas,
			u_n_rtcn,
			u_n_mem,
			u_n_locv,
			u_n_init,
			u_n_modc,
			u_n_func,
			u_n_spipb,
			u_n_symr,
			in_po;
unsigned char		u_f_sim,
			u_f_case;

/* delay management variables */
ODELAY_PTR		*del_con_tab;	/* indirect pointer to the constant
					   delay table */
unsigned long		n_del_con,	/* amount of constant delays */
			u_n_delay,	/* global amount of delays */
			u_del_imm;	/* amount of immediate delays */

unsigned long	        n_nint_con =0;  /* amount of not integer constants */

static STAT_PTR		*deb;

/*************************************************************************

   ROUTINE:     load_unit()

   FUNCTION:    load file .DEO for the specified unit.
                Returns the pointer to the generated structure.

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

OHEADER_PTR load_unit(name, d_name)
char *name;     /* name of the unit to be loaded */
long *d_name;   /* pointer to the field of the unit description which contains
		   the index of the corresponding element in the par_des 
		   structure */

{
/* external variables */

        extern unsigned long    dir_number;
        extern IN_ST            par_des[];
	extern unsigned long	par_pun[];

/* external functions */

        char            *alloca();
        ODELAY_PTR      load_delay();
        STAT_PTR        *load_static();
        SYMT_PTR        load_symtab();
        FILE            *fopen();
	int		fseek();
	char		open_lib();
	IND_PTR		search();

/* local variables */

        long            i;
	long		j;
	char		*path_n = NULL;
	IND_PTR		p;
	IND_PTR		p_aux;

/* code */

        /* create file name : <name>.DEO */
        if (strlen(name) > NAMELENGTH) NAME_ERR(name);
        sprintf(fobj, "%s%s", name, TOP_OBJ);
        /* open the object file .DEO under the current directory */
        if ((f = fopen(fobj, "r")) == NULL)
        {
                /* Search the unit under directory or library if it has been 
		specified */

                for (i = 0; i < dir_number && path_n == NULL; i++)
                {
		   if (par_des[j = par_pun[i]].type == DIR)
		   {
                       /* create file name : directory<name>.DEO */
#if UNIX_WA
sprintf(fobj, "%s/%s%s", par_des[j].name, name,TOP_OBJ);
#else
sprintf(fobj, "%s%s%s", par_des[j].name, name, TOP_OBJ);
#endif
                       /* open file .DEO under directory */
                       if ((f = fopen(fobj, "r")) != NULL)
                            path_n = par_des[*d_name = j].name;
		    }
		    else 
		    {
			   /* open library if not yet opened */
			   if (par_des[j].lib == NULL)
			   {
				par_des[j].lib = (LIB_PTR) alloca(sizeof(LIB));
				if (open_lib(par_des[j].name, par_des[j].lib,
				    READ) == FALSE)
				{
					/* error opening library */
					release(par_des[j].lib, sizeof(LIB));
					par_des[j].lib = NULL;
				}
			    }
			    if (par_des[j].lib != NULL)
			    {
			        /* library opened  - search the unit */
			        if ((p = search(name, par_des[j].lib,
					&p_aux)) != NULL && p->in_d.val_flag)
			    	    {
					f = par_des[j].lib->f;
 			 		if (fseek(f, p->in_d.p_deo, 0) == EOF)
								POS_ERR;
			        	path_n = par_des[*d_name = j].name;
				    }
			    }
		    }
                }
                if (path_n == NULL) fatal2("linker can't load unit", name);
        }
        
        /* read and load header section */
        load_header();
        
        /* create the list of the topological indices of the modules 
           definitions */
        p_ld = NULL;
        if (p_h->n_mod > 0) 
        {
                p_ld = (long*) alloca(p_h->n_mod * sizeof(long));
        }
        p_h->list_def = p_ld;
        p_h->n_inst = p_h->n_umod;
        u_n_spipb = 0;
        
        /* load delay section */
        delay_unit = load_delay();
	u_del_imm = u_n_delay - n_del_con;
	if (n_del_con > 0)
	{
	    del_con_tab = (ODELAY_PTR*) alloca(n_del_con * sizeof(ODELAY_PTR));
	    for (i = 0; i < n_del_con; i++)
		 del_con_tab[i] = (ODELAY_PTR)(delay_unit + u_del_imm + i);
	}
	else
	{
	    del_con_tab = NULL;
	}
    
	/* load modules and functions C section */
	if (u_n_modc > 0) load_unitc();

        /* load topological section */
        p_h->stat_unit = load_static();

        if (u_n_init > 0) load_init();

        /* evaluate the number of symbols referring to units 
           (instantiations and definitions) */
        u_n_symr = p_h->n_mod + p_h->n_umod;
        
        /* Read the number of internal symbols from the
           first record of the symbol table */
        fscanf(f, "%ld %*ld %*ld", &p_h->n_sym);
         

        /* load symbol table for the internal symbols */
        p_h->symb_unit = load_symtab(p_h->n_sym);

        /* load symbol table for the units instant. & definitions */
        p_h->symb_def = (u_n_symr > 0) ? load_symtab(u_n_symr)
                                       : NULL;
        
        /* close the file .DEO */
	if (*d_name == MAX_UNSIGNED || par_des[*d_name].type == DIR)
        	if (fclose(f) != 0) bug2("linker can't close file", fobj);

	if (p_h->unit_type == U_COMPONENT)
	  for (i = 0; i < p_h->n_top; i++) p_h->stat_unit[i]->attr = DUMMY;

        return(p_h);
}


/************************************************************************

   ROUTINE:     load_header

   FUNCTION:    load header section from file .DEO

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES:       Some header informations are skipped.
                26/01/89 (L.Croce): c_lrel changed from 'long' to 'char *'.

   METHOD: 

**********************************************************************/

load_header()

{
/* external functions */

        char    *alloca();
	char	*obj_rel();
	char	*match_rel();

/* local variables */

        char    c_rel[8];
        long    aux_vdd;
        long    aux_gnd;
        long    aux;

/* code */

        /* allocate header structure */
        p_h = (OHEADER_PTR) alloca(sizeof(*p_h));       
        
        /* read from file (sequentially) */
        fscanf(f, "%s", c_rel);

        /* control the compiler release */
	if (strcmp(obj_rel(c_rel), match_rel(C_CREL)) !=0)
	{
	   printf("\n\nWrong compiler release, please use release %s\n\n",
	         match_rel(C_CREL));
	   exit(MOZ_ERR);
	}
 
        fscanf(f, "%ld %*s", &aux);	/* The unit name is already known */

	p_h->unit_type = aux;
        fscanf(f, "%ld", &p_h->warn_cnt);
        fscanf(f, "%ld %ld", &p_h->n_pi, &u_n_bitpi);
        fscanf(f, "%ld %ld", &p_h->n_po, &u_n_bitpo);
        fscanf(f, "%ld %ld", &p_h->n_pb, &u_n_bitpb);
        fscanf(f, "%ld", &u_n_const);
        fscanf(f, "%*ld");      /* nodes */
        fscanf(f, "%*ld %*ld"); /* wires */
        fscanf(f, "%*ld %*ld"); /* registers */
        fscanf(f, "%ld", &u_n_tran);
        fscanf(f, "%ld", &u_n_gate);
        fscanf(f, "%ld", &u_n_rtas);
        fscanf(f, "%ld", &u_n_rtcn);
        fscanf(f, "%ld", &u_n_mem);
        fscanf(f, "%ld", &u_n_locv);
        fscanf(f, "%ld", &aux);    /* for future use */
        fscanf(f, "%ld %ld", &p_h->n_mod, &p_h->n_umod); 
        fscanf(f, "%ld %ld %ld %ld", &u_n_delay, &n_del_con, &p_h->n_top,
		 &u_n_init);
	fscanf(f, "%ld %ld", &aux, &aux); /* for future use */
	fscanf(f, "%ld %ld", &u_n_modc, &u_n_func);
        fscanf(f, "%ld", &aux);
        u_f_sim = aux;
        fscanf(f, "%ld", &aux);
        u_f_case = aux;
        fscanf(f, "%ld %ld %ld", &aux_vdd, &aux_gnd, &in_po);
        p_h->vdd = (aux_vdd < 0) ? MAX_UNSIGNED : aux_vdd;
        p_h->gnd = (aux_gnd < 0) ? MAX_UNSIGNED : aux_gnd;
        fscanf(f, "%*ld");      /* compilation date */
        fscanf(f, "%*ld");      /* compilation time */
        fscanf(f, "%ld", &p_h->c_ctime);
        fscanf(f, "%ld", &p_h->time_unit);
} 

        
/*************************************************************************

   ROUTINE:     load_delay

   FUNCTION:    load delay section from file .DEO.
                Returns start address of the generated array
                storing one delay to each element.
		Immediate delays precede the constant delays

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

ODELAY_PTR load_delay()

{
/* external functions */

        char            *alloca();

/* local variables */

        unsigned long   i;
        ODELAY_PTR      p_d;

/* code */
        
        p_d = (ODELAY_PTR) alloca(u_n_delay * sizeof(*p_d));

        for (i = 0; i < u_n_delay; i++) 
        {
                fscanf(f, "%ld %ld",                       
                        &p_d[i].rise,
                        &p_d[i].fall);

		p_d[i].time_unit = p_h->time_unit;
		p_d[i].rise_type = DEL_IMM;
		p_d[i].fall_type = DEL_IMM;
		if (p_d[i].rise & (unsigned)0x80000000)
		{
			p_d[i].rise_type = DEL_CON;
			p_d[i].rise &= 0x7fffffff;
		}
		if (p_d[i].fall & (unsigned)0x80000000)
		{
			p_d[i].fall_type = DEL_CON;
			p_d[i].fall &= 0x7fffffff;
		}
        }
        return(p_d);
}



/*************************************************************************

   ROUTINE:     load_static

   FUNCTION:    load topological section from file .DEO
                Returns the start address of the array generated 
                with one topological descriptor to each element.
                
   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

STAT_PTR *load_static()

{
/* external functions */

        CONN_PTR        load_conn();
        STAT_PTR        *new_top_el();
	char		*strsave();

/* local variables */

        STAT_PTR        *p_s;
        unsigned long   i;
        unsigned long   j;
        unsigned long   delay;
        char            val[CONSTLENGTH];
        long            aux,
        		aux1,
        		aux2,
			aux3,
			aux4;
	unsigned long	n_fanout,
			n_fanin,
			n_conn,
			n_code;
	long		ch;

/* code */

   /* create an array of topological elements */
   p_s = new_top_el(p_h->n_top, 0);

   for (i = 0; i < p_h->n_top; i++)
   {
        fscanf(f, "%ld %ld %ld", &aux, &aux1, &aux2);
	if (aux != INTERN && aux != DUMMY && aux != PI && aux != PO &&
	    aux != PB) bug("Load: illegal topological attribute");
        p_s[i]->attr = aux;
        p_s[i]->code = aux1;
	if (aux2 && (aux1 != NOD && aux1 != WIR && aux1 != REG && aux1 != GAT))
		bug("Load: illegal role field");
        p_s[i]->function_flag = aux2;

        switch (p_s[i]->code)
        {
           case NOD :
           case WIR :
           case REG :
                fscanf(f, "%ld %ld %ld %ld %ld %ld",
                        &aux,
                        &aux1,
                        &delay,
                        &n_fanout,
                        &n_fanin,
                        &n_conn);
		if (aux > 255) bug("Load: illegal carrier dimension");
		p_s[i]->C.dim = aux;
		if (aux1 > 255) bug("Load: illegal carrier size");
                p_s[i]->C.size = aux1;
                p_s[i]->C.l_fanout = (n_fanout > 0) ?
			load_conn(n_fanout, NULL) : NULL;
                p_s[i]->C.l_fanin = (n_fanin > 0) ? 
			load_conn(n_fanin, NULL) : NULL;
                p_s[i]->C.l_conn = (n_conn > 0) ?
			load_conn(n_conn, NULL) : NULL;
                /* set absolute pointer to corresponding delay if it is
		   immediate - otherwise the offset to the constant delay 
		   table is set */
		if (delay >= u_n_delay) bug("Load: illegal delay definition");
		if (delay < u_del_imm)
		{
                	p_s[i]->C.delay.del_ptr = delay_unit + delay;
                	p_s[i]->C.del_type = DEL_IMM;
		}
		else
		{
                	p_s[i]->C.delay.del_off = delay - u_del_imm;
                	p_s[i]->C.del_type = DEL_CON;
		}

                /* initialize other fields */
                p_s[i]->C.init_el = NULL;
                break;

           case LOCV :
                fscanf(f, "%ld %ld %ld %ld",
                        &aux,
                        &aux1,
                        &n_fanout,
                        &n_fanin);
		if (aux1 > 255) bug("Load: illegal carrier dimension");
		p_s[i]->C.dim = aux1;
		p_s[i]->C.del_type = aux;
                p_s[i]->C.l_fanout = (n_fanout > 0) ?
			load_conn(n_fanout, NULL) : NULL;
                p_s[i]->C.l_fanin = (n_fanin > 0) ? 
			load_conn(n_fanin, NULL) : NULL;

                /* initialize other fields */
                p_s[i]->C.init_el = NULL;
                break;
           case TRAN :
                fscanf(f, "%ld %ld", &aux, &aux1);
		if (aux != _DEPLETION && aux != _NCHANNEL && aux != _PCHANNEL)
		   bug("Load: Illegal transistor type");
		if (aux1 > 15) bug("Load: illegal strength value");
                p_s[i]->T.type = aux;
                p_s[i]->T.str = aux1;
                p_s[i]->T.drain = load_conn(1, NULL);
                p_s[i]->T.gate = load_conn(1, NULL);
                p_s[i]->T.source = load_conn(1, NULL);
                break;
           case GAT :
                fscanf(f, "%ld %ld %ld %ld %ld %ld",
                        &aux,
                        &delay,
                        &aux1,
                        &aux2,
                        &n_fanout,
                        &n_fanin);
		/* test sul tipo di gate */
                p_s[i]->G.type = aux;
		if (aux1 > 15 || aux2 > 15) bug("Load: illegal strength value");
                p_s[i]->G.str_0 = aux1;
                p_s[i]->G.str_1 = aux2;
                p_s[i]->G.l_fanout = (n_fanout > 0) ?
			load_conn(n_fanout, NULL) : NULL;
                p_s[i]->G.l_fanin = (n_fanin > 0) ?
			load_conn(n_fanin, NULL) : NULL;
                /* set absolute pointer to corresponding delay if it is
		   immediate - otherwise the offset to the constant delay 
		   table is set */
		if (delay >= u_n_delay) bug("Load: illegal delay definition");
		if (delay < u_del_imm)
		{
                	p_s[i]->G.delay.del_ptr = delay_unit + delay;
                	p_s[i]->G.del_type = DEL_IMM;
		}
		else
		{
                	p_s[i]->G.delay.del_off = delay - u_del_imm;
                	p_s[i]->G.del_type = DEL_CON;
		}

                /* initialize other fields */
                p_s[i]->G.init_el = NULL;
                p_s[i]->G.s_lst = NULL;
                break;
           case DEF :
                fscanf(f, "%ld", &n_fanout);
                p_s[i]->D.l_ist = (n_fanout > 0) ? 
			load_conn(n_fanout, NULL) : NULL;

                /* insert the element index into the unit definition list*/
                *p_ld++ = i;
                break;
           case USE :
                fscanf(f,"%ld %ld %ld",
                        &n_fanout,
                        &n_fanin,
                        &n_conn);
                p_s[i]->U.l_fanout = (n_fanout > 0) ?
			load_conn(n_fanout, NULL) : NULL;
                p_s[i]->U.l_fanin = (n_fanin > 0) ?
			load_conn(n_fanin, NULL) : NULL;
                p_s[i]->U.l_fanbin = (n_conn > 0) ?
			load_conn(n_conn, NULL) : NULL;
                break;
           case RT_AS :
           case RT_LD :
                fscanf(f, "%ld %ld %ld %ld %ld %ld %ld",
			&aux,
                        &delay,
                        &aux1,
                        &aux2,
                        &n_fanout,
                        &n_fanin,
                        &n_code);
		if (aux > 255) bug("Load: illegal rt dimension");
                p_s[i]->A.bit = aux;
		if (aux1 > 15 || aux2 > 15) bug("Load: illegal strength value");
                p_s[i]->A.str_0 = aux1;
                p_s[i]->A.str_1 = aux2;
                p_s[i]->A.l_fanout = (n_fanout > 0) ?
			load_conn(n_fanout, NULL) : NULL;
                p_s[i]->A.l_fanin = (n_fanin > 0) ?
			load_conn(n_fanin, NULL) : NULL;
                p_s[i]->A.l_code = (n_code > 0) ?
			load_conn(n_code, CODE) : NULL;
                /* set absolute pointer to corresponding delay if it is
		   immediate - otherwise the offset to the constant delay 
		   table is set */
		if (delay >= u_n_delay) bug("Load: illegal delay definition");
		if (delay < u_del_imm)
		{
                	p_s[i]->A.delay.del_ptr = delay_unit + delay;
                	p_s[i]->A.del_type = DEL_IMM;
		}
		else
		{
                	p_s[i]->A.delay.del_off = delay - u_del_imm;
                	p_s[i]->A.del_type = DEL_CON;
		}

                /* initialize other fields */
                break;  
           case RT_CN :
           case RT_CNT :
                fscanf(f, "%ld %ld %ld",
                        &n_fanout,
                        &n_fanin,
                        &n_code);
                p_s[i]->O.l_fanout = (n_fanout > 0) ?
			load_conn(n_fanout, NULL) : NULL;
                p_s[i]->O.l_fanin = (n_fanin > 0) ?
			load_conn(n_fanin, NULL) : NULL;
                p_s[i]->O.l_code = (n_code > 0) ?
			load_conn(n_code, CODE) : NULL;
                break;
           case CONST :
                fscanf(f, "%ld %ld %ld %ld %ld ", 
			&aux,
			&n_fanout,
			&n_fanin,
			&n_conn,
                        &n_code);
		j = 0;
		while ((ch = fgetc(f)) != '\n') val[j++] = ch;
		val[j] = '\0';
		p_s[i]->I.type = aux;
		p_s[i]->I.leng = strlen(val);
                p_s[i]->I.val = strsave(val);
		switch (p_s[i]->I.type)
		{
		   case C_WIRE :
                        n_nint_con++;
			/* these constants may have fanout and connectivity */
			if (n_fanin > 0 || n_code > 0)
				bug("Load: illegal wire constant");
	                p_s[i]->IW.l_fanout = (n_fanout > 0) ?
				load_conn(n_fanout, NULL) : NULL;
	                p_s[i]->IW.l_conn = (n_conn > 0) ?
				load_conn(n_conn, NULL) : NULL;
	                break;
		   case C_INT :
			/* these constants may have fanout, fanin and code */
			if (n_conn > 0)
				bug("Load: illegal int constant");
	                p_s[i]->ID.l_fanout = (n_fanout > 0) ?
				load_conn(n_fanout, NULL) : NULL;
    	            	p_s[i]->ID.l_fanin = (n_fanin > 0) ?
				load_conn(n_fanin, NULL) : NULL;
	                p_s[i]->ID.l_code = (n_code > 0) ?
				load_conn(n_code, CODE) : NULL;
			p_s[i]->ID.evval = atoi(p_s[i]->I.val);
			break;
		   case C_CHAR :
                        n_nint_con++;
			/* these constants may have only fanout */
			if (n_fanin > 0 || n_code > 0 || n_conn > 0)
				bug("Load: illegal character constant");
	                p_s[i]->IC.l_fanout = (n_fanout > 0) ?
				load_conn(n_fanout, NULL) : NULL;
			break; 
		   default :
			bug("Load: illegal constant type");
			break;
		}
                break;
           case MEM:
                fscanf(f, "%ld %ld %ld %ld %ld %ld %ld %s",
                        &aux,
                        &aux1,
                        &aux2,
			&aux3,
			&aux4,
                        &delay,
                        &n_fanin,
                        val);
                p_s[i]->M.type = aux;
		if (aux1 > 255) bug("Load: illegal word size");
                p_s[i]->M.word_size = aux1;
		if (aux2 > 255) bug("Load: illegal address size");
                p_s[i]->M.addr_size = aux2;
		if (aux3 > 15 || aux4 > 15) bug("Load: illegal strength value");
                p_s[i]->M.str_0 = aux3;
                p_s[i]->M.str_1 = aux4;
                p_s[i]->M.l_fanout = load_conn(1, NULL);
                p_s[i]->M.l_fanin = (n_fanin > 0) ?
			load_conn(n_fanin, NULL) : NULL;
                p_s[i]->M.val = strsave(val);
                /* set absolute pointer to corresponding delay if it is
		   immediate - otherwise the offset to the constant delay 
		   table is set */
		if (delay >= u_n_delay) bug("Load: illegal delay definition");
		if (delay < u_del_imm)
		{
                	p_s[i]->M.delay.del_ptr = delay_unit + delay;
                	p_s[i]->M.del_type = DEL_IMM;
		}
		else
		{
                	p_s[i]->M.delay.del_off = delay - u_del_imm;
                	p_s[i]->M.del_type = DEL_CON;
		}

                /* initialize other fields */
                break;
  	   case AS_TR :
                fscanf(f, "%hd %ld %ld %ld %ld %ld %ld",
                        &p_s[i]->A.bit,
                        &aux2,
	                &aux,
                	&aux1,
                        &n_fanout,
                        &n_fanin,
                        &n_code);
                p_s[i]->A.type = aux2;
                p_s[i]->A.str_0 = aux;
                p_s[i]->A.str_1 = aux1;
                p_s[i]->A.l_fanout = (n_fanout > 0)
				? load_conn(n_fanout, NULL) : NULL;
                p_s[i]->A.l_fanin = (n_fanin > 0)
				? load_conn(n_fanin, NULL) : NULL;
                p_s[i]->A.l_code = (n_code > 0) 
				? load_conn(n_code, CODE) : NULL;
                break;

           case PROC :
                fscanf(f, "%ld %ld %ld %ld %ld %ld",
                        &aux2,
                        &n_fanout,
                        &aux,
                        &n_fanin,
                        &aux1,
                        &n_code);
                p_s[i]->P.type = aux2;
                p_s[i]->P.l_fanout = (n_fanout > 0)
				? load_conn(n_fanout, NULL) : NULL;
                p_s[i]->P.l_tfanin = (aux > 0)
				? load_conn(aux, NULL) : NULL;
                p_s[i]->P.l_dfanin = (n_fanin > 0)
				? load_conn(n_fanin, NULL) : NULL;
                p_s[i]->P.l_locvar = (aux1 > 0)
				? load_conn(aux1, NULL) : NULL;
                p_s[i]->P.l_code = (n_code > 0) 
				? load_conn(n_code, CODE) : NULL;
                break;  
	default :
		bug("Element not managed");
		break;
        }
   }
   return(p_s);
}


/*************************************************************************

   ROUTINE:     load_conn

   FUNCTION:    Read the variable part of a topological descriptor.
                Returns the start pointer of the generated list.

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 	All the lists are composed by 4 fields

   METHOD: 

**********************************************************************/

CONN_PTR load_conn(n_el, list_type)
unsigned long n_el;        /* number of elements to be read */
char list_type;   /* indicates the type of the elements to be read 
                    in order to distinguish the computational appendix
                    from the other kinds of elements */
{
/* external functions */

        CONN_PTR        new_c_el();
        CONN_PTR        first_c();

/* local variables */

static CONN_PTR		p_conn;
static long		aux_type;
static long		signal;
static long		first;
static long		second;

/* code */

        /* initialize a new list of type CONN */
        new_c_list();

        while (n_el-- > 0)
        {
                /* create a new element of type CONN */
                fscanf(f, "%ld %ld %ld %ld",
			&aux_type,
			&signal,
                        &first,
                        &second);
                p_conn = new_c_el(aux_type, signal, first, second);

		if (list_type == CODE && p_conn->type == CALL)
			p_conn->CL.fun_p = p_conn->signal + func_unit;

		/* gestire delay */
		if (list_type == CODE && p_conn->type == TIME)
		{
		   if (p_conn->signal < u_del_imm)
		   {	
			p_conn->CT.del_p.del_ptr = p_conn->signal + delay_unit;
                	p_conn->signal = DEL_IMM;
		   }
		   else
		   {
			p_conn->CT.del_p.del_off = p_conn->signal - u_del_imm;
                	p_conn->signal = DEL_CON;
		   }
		}


                /* link to list */
                enqueue_c(p_conn);
        }
        return(first_c());
}


/*************************************************************************

   ROUTINE:     load_symtab

   FUNCTION:    load symbol table

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

SYMT_PTR load_symtab(n_el)
unsigned long n_el;     /* number of elements to be read */

{
/* external functions */

        char            *alloca();
	char		*strsave();

/* local variables */

        unsigned long   el;
        unsigned long   i;
        SYMT_PTR        p_t;
	long		aux;
	char		aux_name[NAMELENGTH + 1];

/* code */
        
        p_t = (SYMT_PTR) alloca(n_el * sizeof(*p_t));

        for (i = 0; i < n_el; i++)
        {
                fscanf(f, "%ld %ld %s", &p_t[i].st_el, &aux, aux_name);

		p_t[i].el_type = aux;
		p_t[i].el_name = strsave(aux_name);

                /* set pointer (absolute) from static element to symbol table */
                el = p_t[i].st_el;

		/* Gates may have two names when they are integrated with
		   wires. For gates PO we have to store all the names
		   in order to correctly create the interface tables.
		   The first name is pointed by s_tab, the second one 
		   is pointed by the s_lst pointer. */
                if (p_h->stat_unit[el]->attr == PO &&
		    p_h->stat_unit[el]->code == GAT)
		{
		   if (p_h->stat_unit[el]->s_tab == NULL)
			/* first name */
			p_h->stat_unit[el]->s_tab = p_t + i;
		   else
			/* second name */
		     	p_h->stat_unit[el]->G.s_lst = p_t + i;
		}
		else
			/* element with only a name */
                	p_h->stat_unit[el]->s_tab = p_t + i;

                /* increment counter of PO/PI/PB symbols */
                if (p_h->stat_unit[el]->attr != INTERN &&
		    p_h->stat_unit[el]->code != CONST) u_n_spipb++;
        }
        return(p_t);
}


/*************************************************************************

   ROUTINE:     load_init

   FUNCTION:    Load the initialization table from file .DEO

   AUTHOR(s):   Daniela Magnani

   DATE:        01/10/1986

   NOTES: 

   METHOD: 

**********************************************************************/

load_init()

{

/* external functions */

	char		*strsave();

/* local variables */

        unsigned long   i;
        unsigned long   el;
        unsigned long   n_fanout;
        char            *c_ptr;
        char            val[CONSTLENGTH];

/* code */
        
        for (i = 0; i < u_n_init; i++)
        {
                fscanf(f, "%s %ld", val, &n_fanout);
                c_ptr = strsave(val);
                while (n_fanout-- > 0)
                {
                        fscanf(f, "%ld", &el);
                        switch (p_h->stat_unit[el]->code)
                        {
                                case WIR :
                                case NOD :
                                case REG :
                                case LOCV :
                                        p_h->stat_unit[el]->C.init_el = c_ptr;
                                        break;
                                case GAT :
                                        p_h->stat_unit[el]->G.init_el = c_ptr;
                                        break;
                                default :
                                        fatal("Illegal initialization");
                                        break;
                        }
                }
        }
}


/************************************************************************

   ROUTINE:     load_unitc

   FUNCTION:    load modules and functions C tables 

   AUTHOR(s):   Daniela Magnani

   DATE:        02/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

load_unitc()

{
/* external functions */

        char            *alloca();
	char		*strsave();

/* local variables */

        unsigned long   i;
        FUNC_PTR	p_f;
	MODC_PTR	p_m;
	unsigned long	mod;
	char		aux_name[NAMELENGTH + 1];

/* code */
        
        modc_unit = p_m = (MODC_PTR) alloca(u_n_modc * sizeof(*p_m));

        for (i = 0; i < u_n_modc; i++) 
        {
                fscanf(f, "%s %*ld %*ld %*ld", aux_name);
		p_m->name = strsave(aux_name);
		p_m++;
        }

	func_unit = p_f = (FUNC_PTR) alloca(u_n_func * sizeof(*p_f));

        for (i = 0; i < u_n_func; i++) 
        {
                fscanf(f, "%ld %ld %ld %ld",
			&p_f[i].func,
			&p_f[i].n_par,
			&mod,
			&p_f[i].res_dim);
		p_f[i].p_mod = mod + modc_unit;
        }
}


/*************************************************************************

   ROUTINE:     new_top_el

   FUNCTION:    create an array of pointers to topological elements.
		The topological elements are created also and initialized.

   AUTHOR(s):   Daniela Magnani

   DATE:        14/07/1988

   NOTES: 

   METHOD: 

**********************************************************************/

STAT_PTR *new_top_el(n_el, start_el)
unsigned long n_el;     /* dimension of the array to be generated */
unsigned long start_el; /* topological descriptor starting element */
{
/* external functions */

        char            *alloca();
	STAT_PTR	new_top_des();

/* local variables */

	STAT_PTR	*p_p_s;
	int		i;

/* code */
        
        p_p_s = (STAT_PTR*) alloca(n_el * sizeof(STAT_PTR));

	for (i = 0; i < start_el; i++) p_p_s[i] = NULL;

	for (i = start_el; i < n_el; i++) p_p_s[i] = new_top_des();

        return(p_p_s);
}


/*************************************************************************

   ROUTINE:     new_top_des

   FUNCTION:    create and initialize a topological descriptor

   AUTHOR(s):   Daniela Magnani

   DATE:        14/07/1988

   NOTES: 

   METHOD: 

**********************************************************************/

STAT_PTR new_top_des()

{
/* external functions */

        char            *alloca();

/* local variables */

static STAT_PTR		p_stat;

/* code */
        
	if (free_stat_list == NULL) create_stat_block();

	p_stat = free_stat_list;
	free_stat_list = p_stat->udesc.free_link;
	free_stat_cnt--;

        p_stat->attr = ILLEGAL;
        p_stat->function_flag = F_NO_FUNCTION;
        p_stat->path_ptr = 0;
        p_stat->pt_el = MAX_UNSIGNED;
        p_stat->s_tab = NULL;
	p_stat->udesc.free_link = NULL;
        return(p_stat);
}


/*************************************************************************

   ROUTINE:     new_ev_top_el

   FUNCTION:    create an array of pointers to the eval portion of the
		topological elements.
		The topological elements are created also and initialized.

   AUTHOR(s):   Daniela Magnani

   DATE:        14/07/1988

   NOTES: 

   METHOD: 

**********************************************************************/

STAT_EV_PTR *new_ev_top_el(n_el)
unsigned long n_el;     /* dimension of the array to be generated */
{
/* external functions */

        char            *alloca();
	STAT_EV_PTR	new_ev_top_des();

/* local variables */

	STAT_EV_PTR	*p_p_s;
	int		i;

/* code */
        
        p_p_s = (STAT_EV_PTR*) alloca(n_el * sizeof(STAT_EV_PTR));

	for (i = 0; i < n_el; i++) p_p_s[i] = new_ev_top_des();

        return(p_p_s);
}


/*************************************************************************

   ROUTINE:     new_ev_top_des

   FUNCTION:    create and initialize the portion of a topological descriptor
		useful only in the eval step

   AUTHOR(s):   Daniela Magnani

   DATE:        14/07/1988

   NOTES: 

   METHOD: 	This element has the same size of the static elements.
		The free_stat_list is then used in order to find elements.

**********************************************************************/

STAT_EV_PTR new_ev_top_des()

{
/* external functions */

        char            *alloca();

/* local variables */

	static  STAT_EV_PTR     p_s;
	static	STAT_PTR	p_stat;

/* code */
        
	if (free_stat_list == NULL) create_stat_block();

	p_stat = free_stat_list;
	free_stat_list = free_stat_list->udesc.free_link;
	free_stat_cnt--;

	p_stat->udesc.free_link = NULL;

	p_s = (STAT_EV_PTR)p_stat;

	p_s->pdelay = NULL;
        p_s->sim_lev = 0;
        p_s->atpg_lev = 0;
        p_s->visited = FALSE;
        p_s->on_path = FALSE;
        p_s->comp = 0;
	p_s->n_fanout = 0;
	p_s->n_fanin = 0;
	p_s->n_conn = 0;
	p_s->n_code = 0;

        return(p_s);
}


/*************************************************************************

   ROUTINE:     free_static

   FUNCTION:    Free a static element (start pointer is given as function 
		parameter)
		
   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 	The space freed is linked to the static free list

   METHOD: 

**********************************************************************/

free_static(p)
STAT_PTR p;

{

/* code */

	p->udesc.free_link = free_stat_list;
	free_stat_list = p;
	free_stat_cnt++;
}


/*************************************************************************

   ROUTINE:     create_stat_block

   FUNCTION:    create a block of static elements
		
   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 	The block is linked to the free stat list.
		The block size is 37 * 512 = 18944

   METHOD:

**********************************************************************/

create_stat_block()
{

/* external functions */

	char	*alloca();

/* local variables */

	char	*p;
	int	i;

/* code */

	block_stat_cnt++;
	p = alloca(NSTAT * DIMSTAT);
	for (i = 0; i < ((NSTAT - 1) * DIMSTAT); i += DIMSTAT)
		((STAT_PTR)((unsigned char *)p + i))->udesc.free_link = 
		 (STAT_PTR)((unsigned char *)p + i + DIMSTAT);

	((STAT_PTR)((unsigned char *)p + i))->udesc.free_link = free_stat_list;
	free_stat_cnt += NSTAT;
	free_stat_list = (STAT_PTR)p;
}

