/* LANGUAGE:    C        VAX VMS COMPILER VERSION 2.0-003


 PACKAGE NAME:          LMAP

 PACKAGE ROUTINES:      cre_map()
                        open_map()
                        head_print()
                        cnt_ln()
                        close_map()
                        loop_map_s()
                        loop_map_d()
                        start_map()
			warnf()

 PACKAGE FUNCTION:      Generate the file .DEM

 PACKAGE AUTHOR(s):     Daniela Magnani

 PACKAGE CREATION DATE: 20/08/1986

**********************************************************************/

# include <stdio.h>
# include "ldef.h"
# include "lstruct.h"

/***************************************************************************/
/*                                                                         */
/*                          VARIABLES DECLARATION                          */
/*                                                                         */
/***************************************************************************/

FILE    *f_m;

static char     res_date[26];
static long     n_page;
static long     n_line;
static char     fmap[NAMELENGTH + 5];


/*************************************************************************

   ROUTINE:     cre_map

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

cre_map()

{

/* external variables */

        extern EHEADER          h_exe;
        extern UNIT_PTR         p_level[];
        extern long             max_level;
	extern IN_ST		par_des[];

/* external functions */

        FILE    *fopen();
        char    *strcat();
        char    *strcpy();

/* local variables */

        UNIT_PTR        u;
        char            ld_date[26];
        long            level;
        long            i;

/* code */

        for (level = max_level; level >= 0; level--)
        {
                for (u = p_level[level]; u != NULL; u = u->level_next)
                {
                        cnt_ln();
                        fprintf(f_m,"\n");
                        n_line++;
                        cnt_ln();
			switch(u->head->unit_type)
			{
				case U_UNIT :
				  fprintf(f_m, "unit : %s ", u->name);
				  break;
				case U_COMPONENT :
				  fprintf(f_m, "component : %s ", u->name);
				  break;
				case U_PROCESS :
				  fprintf(f_m, "process : %s ", u->name);
				  break;
			}
                        if (u->dir != MAX_UNSIGNED)
                        	fprintf(f_m, " (%s)", par_des[u->dir].name);
			fprintf(f_m, "\n");
                        n_line++;
                        time_date(u->head->c_ctime, ld_date);
                        cnt_ln();
                        fprintf(f_m,"compilation date : %s\n", ld_date);
                        n_line++;
                        cnt_ln();
                        fprintf(f_m, "compilation warnings : %d\n",
                                                u->head->warn_cnt);
                        n_line++;
                        for (i = 0; i < u->head->n_umod; i++)
                        {
                                cnt_ln();
                                fprintf(f_m, "   %-18.15s  %-18s\n",
                                        u->par_table[i].in_name,
                                        u->par_table[i].def_name);
                                n_line++;
                        }
                }
        }
}


/*************************************************************************

   ROUTINE:     open_map

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        13/10/1986

   NOTES: 

   METHOD: 

**********************************************************************/

open_map(name)
char *name;

{

/* external variables */

        extern EHEADER          h_exe;

/* external functions */

        FILE    *fopen();
        char    *strcat();
        char    *strcpy();
        int     strlen();

/* code */

        n_page = 1;
        n_line = 0;
        
        if (strlen(name) > NAMELENGTH) NAME_ERR(name);
        strcat(strcpy(fmap, name), MAP_EXE);
        /* open the map file .DEM */
        if ((f_m = fopen(fmap, "w")) == NULL) 
		bug2("linker can't open file", fmap);

        time_date(h_exe.l_ctime, res_date);
        
        head_print();
}

        
/*************************************************************************

   ROUTINE:     head_print()

   FUNCTION:    print on file .DEM headline.

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

head_print()

{
/* external functions */


/* code */

        fprintf(f_m, "%s ver. %s  %s (%s) -  %s  page %d\n",
                        TL_NAME, TL_REL, TL_DATE, TL_CPU, res_date, n_page);
        n_line++;
        fprintf(f_m, "\n");
        n_line++;
}


/*************************************************************************

   ROUTINE:     cnt_ln

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

cnt_ln()

{
/* code */

        if (n_line == MAX_LINES)
        {
                n_line = 0;
                n_page++;
                head_print();
        }
}


/*************************************************************************

   ROUTINE:     close_map

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

close_map()

{
/* external functions */

        int     fclose();

/* code */

        if (fclose(f_m) != 0) bug2("linker can't close file", fmap);
}


/*************************************************************************

   ROUTINE:     loop_map_s

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

loop_map_s(s)
char *s;

{
/* external functions */


/* code */

	n_line++;
        cnt_ln();
        fprintf(f_m,"%s\n", s);
}       


/*************************************************************************

   ROUTINE:     loop_map_d

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

loop_map_d(el)
long el;

{
/* external functions */


/* code */

        n_line++;
	cnt_ln();
        fprintf(f_m, "%d\n", el);
}


/*************************************************************************

   ROUTINE:     start_map

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        20/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

start_map(s)
char *s;

{
/* external functions */


/* code */

	n_line++;
        cnt_ln();
        fprintf(f_m, "\n%s on path :\n", s);
}


/*************************************************************************

   ROUTINE:     warnf

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        28/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

warnf(s)
char *s;

{
/* external functions */


/* code */

	n_line++;
        cnt_ln();
        fprintf(f_m, "\noutput signal used as inout: %s\n", s);
}

