/* LANGUAGE:    C        VAX VMS COMPILER VERSION 2.0-003


 PACKAGE NAME:          LUTIL

 PACKAGE ROUTINES:      fatal()
                        fatal2()
                        bug()
			bug2()
			prog_bug()
			warn()
                        warn2()
			warn2d()
                        date_and_hour()
                        to_up()
                        to_low()
                        time_date()
                        quicksort()
			strsave()
			alloca()
			release()
			obj_rel()
			match_rel()
			
 PACKAGE FUNCTION:      utility functions

 PACKAGE AUTHOR(s):     Daniela Magnani

 PACKAGE CREATION DATE: 19/08/1986

**********************************************************************/

# include <time.h>
# include <ctype.h>
# include <stdio.h>
# include "ldef.h"
# include "lstruct.h"


/*************************************************************************

   ROUTINE:     fatal()

   FUNCTION:    signal fatal error and abort the linker.
                The program is exited with code MOZ_ERR (1).

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

fatal(s)
char *s;        /* error message */

{
/* external variables */

        extern  FILE    *f_m;

/* code */

        printf("\nFATAL : %s\nlinker aborted\n", s);
        STOP;
	if (f_m != NULL)
	{
        	fprintf(f_m, "\nFATAL : %s\nlinker aborted\n", s);
        	fclose(f_m);
	}
        exit(MOZ_ERR);
}


/*************************************************************************

   ROUTINE:     fatal2()

   FUNCTION:    signal fatal error and abort the linker.
                The program is exited with code MOZ_ERR (1).

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

fatal2(s, s1)
char *s, *s1;   /* error messages */

{
/* external variables */

        extern FILE     *f_m;

/* code */

        printf("\nFATAL : %s %s\nlinker aborted\n", s, s1);
        STOP;
	if (f_m != NULL)
	{
        	fprintf(f_m, "\nFATAL : %s %s\nlinker aborted\n", s, s1);
        	fclose(f_m);
	}
        exit(MOZ_ERR);
}



/*************************************************************************

   ROUTINE:     bug()

   FUNCTION:    signal bug (system error )and abort the linker.
                The program is exited with code MOZ_SYS (2).

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

bug(s)
char *s;        /* error message */

{
/* external variables */

        extern  FILE    *f_m;

/* code */

        printf("\nSYSTEM ERROR : %s\nlinker aborted\n", s);
        STOP;
	if (f_m != NULL)
	{
        	fprintf(f_m, "\nSYSTEM ERROR : %s\nlinker aborted\n", s);
        	fclose(f_m);
	}
        exit(MOZ_SYS);
}


/*************************************************************************

   ROUTINE:     bug2()

   FUNCTION:    signal bug (system error) and abort the linker.
                The program is exited with code MOZ_SYS (2).

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

bug2(s, s1)
char *s, *s1;   /* error messages */

{
/* external variables */

        extern FILE     *f_m;

/* code */

        printf("\nSYSTEM ERROR : %s %s\nlinker aborted\n", s, s1);
        STOP;
	if (f_m != NULL)
	{
        	fprintf(f_m, "\nSYSTEM ERROR : %s %s\nlinker aborted\n", s, s1);
        	fclose(f_m);
	}
        exit(MOZ_SYS);
}



/*************************************************************************

   ROUTINE:     prog_bug()

   FUNCTION:    signal bug (program) and abort the linker.
                The program is exited with code MOZ_BUG (3).

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

prog_bug(s)
char *s;        /* error message */

{
/* external variables */

        extern  FILE    *f_m;

/* code */

        printf("\nPROGRAM BUG : %s\nlinker aborted\n", s);
        STOP;
	if (f_m != NULL)
	{
	        fprintf(f_m, "\nPROGRAM BUG : %s\nlinker aborted\n", s);
       	 	fclose(f_m);
	}
        exit(MOZ_BUG);
}

/*************************************************************************

   ROUTINE:     warn()

   FUNCTION:    Display warning message.

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

warn(s)
char *s;        /* warning message */

{
/* external variables */

        extern FILE     *f_m;

/* code */
        printf("\nwarning : %s\n", s);
        fprintf(f_m, "\nwarning : %s\n", s);
}


/*************************************************************************

   ROUTINE:     warn2()

   FUNCTION:    Display warning message.

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

warn2(s, s1)
char *s;        /* warning message */
char *s1;

{
/* external variables */

        extern FILE     *f_m;

/* code */
        printf("\nwarning : %s %s\n", s, s1);
        fprintf(f_m, "\nwarning : %s %s\n", s, s1);
}


/*************************************************************************

   ROUTINE:     warn2d()

   FUNCTION:    Display warning message followed by an integer value.

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

warn2d(d, s)
long d;
char *s;        /* warning message */

{
/* external variables */

        extern FILE     *f_m;

/* code */
        printf("\nwarning : %d %s\n", d, s);
        fprintf(f_m, "\nwarning : %d %s\n", d, s);
}


/*************************************************************************

   ROUTINE:     date_and_hour()

   FUNCTION:    converts the time vtime (elapsed since 00:00:00, January
                1, 1970 in seconds) in date (ddmmyyyy) and hour (hhmmss)

   AUTHOR(s):   Antonio Lioy

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

date_and_hour(vtime, day, hour)
long vtime;     /* time elapsed since 00:00:00, January 1, 1970
                           in seconds */
long *day;      /* pointer to resulting date */
long *hour;     /* pointer to resulting hour */

{

/* external functions */

        char    *ctime();
        char    *strcpy();
        int     strcmp();

/* local variables */

        char    temp[26];
        char    temp1[4];
        long    i;              

/* code */

         strcpy(temp, ctime(&vtime));
         temp1[0] = temp[4];
         temp1[1] = temp[5];
         temp1[2] = temp[6];
         temp1[3] = '\0';
         
         /* day  */
         temp[0] = temp[8];
         temp[1] = temp[9];
         if(temp[0] == ' ') temp[0] = '0'; 

         /* month */
         temp[2] = '0';
         if (strcmp(temp1,"Jan") == 0) temp[3] = '1';
         if (strcmp(temp1,"Feb") == 0) temp[3] = '2';
         if (strcmp(temp1,"Mar") == 0) temp[3] = '3';
         if (strcmp(temp1,"Apr") == 0) temp[3] = '4';
         if (strcmp(temp1,"May") == 0) temp[3] = '5';
         if (strcmp(temp1,"Jun") == 0) temp[3] = '6';
         if (strcmp(temp1,"Jul") == 0) temp[3] = '7';
         if (strcmp(temp1,"Aug") == 0) temp[3] = '8';
         if (strcmp(temp1,"Sep") == 0) temp[3] = '9';
         if (strcmp(temp1,"Oct") == 0) {temp[3] = '0' ; temp[2] = '1';}
         if (strcmp(temp1,"Nov") == 0) {temp[3] = '1'; temp[2] = '1';}
         if (strcmp(temp1,"Dec") == 0) {temp[3] = '2'; temp[2] = '1';}

         /* year */
         temp[4] = temp[20];
         temp[5] = temp[21];
         temp[6] = temp[22];
         temp[7] = temp[23];

         temp[8] = temp[11];
         temp[9] = temp[12];

         temp[10] = temp[14];
         temp[11] = temp[15];

         temp[12] = temp[17];
         temp[13] = temp[18];

         for (i = 0; i < 14; i++) temp[i] = temp[i] - '0';
 
         *day = NULL;
         for (i = 0; i < 8; i++) *day = *day * 10 + temp[i];
                 
         *hour = NULL;
         for (i = 8; i < 14; i++) *hour = *hour * 10 + temp[i];
} 


/*************************************************************************

   ROUTINE:     to_up()

   FUNCTION:    converts a string to uppercase.

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

char *to_up(st)
char *st;
{
/* local variables */

	char	*s;

/* code */
	
	s = st;
        while (*s != NULL)
        {
                *s = TOUPPER(*s);
                s++;
        }
	return(st);
}


/*************************************************************************

   ROUTINE:     to_low()

   FUNCTION:    converts a string to lowercase

   AUTHOR(s):   Daniela Magnani

   DATE:        30/03/1987

   NOTES:

   METHOD: 

**********************************************************************/

to_low(s)
char *s;
{
/* code */

        while (*s != NULL)
        {
                *s = TOLOWER(*s);
                s++;
        }
}



/*************************************************************************

   ROUTINE:     time_date()

   FUNCTION:    returns the time value vtime converted to an ASCII
                string of the form: Wkd dd mmm 19yy hh:mm:ss\0

   AUTHOR(s):   Daniela Magnani

   DATE:        19/08/1986

   NOTES:

   METHOD: 

**********************************************************************/

time_date(vtime, temp1)
long vtime;     /* time in seconds, since 00:00:00 January 1,1970 */
char *temp1;    /* output string  - 26 characters */

{

/* external functions */

        char    *ctime();
        char    *strcpy();
        
/* local variables */

        char    temp[26];

/* code */

        strcpy(temp, ctime(&vtime));
        strcpy(temp1, temp);
                
        /* dd */
        temp1[4] = temp[8];
        temp1[5] = temp[9];
        temp1[6] = temp[10];

        /* mmm */
        temp1[7] = temp[4];
        temp1[8] = temp[5];
        temp1[9] = temp[6];
        temp1[10] = temp[7];

        /* 19yy */
        temp1[11] = temp[20];
        temp1[12] = temp[21];
        temp1[13] = temp[22];
        temp1[14] = temp[23];
        temp1[15] = temp[19];

        /* hh:mm:ss */
        temp1[16] = temp[11];
        temp1[17] = temp[12];
        temp1[18] = temp[13];
        temp1[19] = temp[14];
        temp1[20] = temp[15];
        temp1[21] = temp[16];
        temp1[22] = temp[17];
        temp1[23] = temp[18];

        temp1[24] = temp[25];
}
        

/*************************************************************************

   ROUTINE:     quicksort

   FUNCTION:    #

   AUTHOR(s):   Antonio Lioy

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

quicksort(l, r)
INTF_PTR *l;
INTF_PTR *r;

{
/* external functions */

        int             strcmp();

/* local variables */

        INTF_PTR        *i;
        INTF_PTR        *j;
        char            *check_name;

/* code */

        i = l; j = r; check_name = (*(l + (r - l)/2))->name;
        do
        {
                while (strcmp((*i)->name, check_name) < 0) i++;
                while (strcmp((*j)->name, check_name) > 0) j--;
                if (i <= j)
                {
                        if (i < j)
                        {
                                INTF_PTR temp = *i;
                                *i = *j;
                                *j = temp;
                        }
                        i++;
                        j--;
                }
        } while ( i <= j);
        if (l < j) quicksort(l, j);
        if (i < r) quicksort(i, r);
}


/*************************************************************************

   ROUTINE:     moz_clock

   FUNCTION:    Return the cpu-time (in 10 millisecond unit) elapsed from
                the value of the old_time parameter.

   AUTHOR(s):   Antonio Lioy

   DATE:        17/04/1987

   NOTES: 

   METHOD: 

**********************************************************************/

long moz_clock(old_time)
long old_time;

{
/* local variables */

        struct {
        int proc_user;
        int proc_system;
        int child_user;
        int child_system;
        } actual;

/* body */

        /* Times is called to fill actual */

        times (&actual);

        /* Elapsed user process time is computed */

        return (actual.proc_user - old_time);
}


/*************************************************************************

   ROUTINE:     strsave

   FUNCTION:    Allocate a string.

   AUTHOR(s):   Daniela Magnani

   DATE:        08/07/1988

   NOTES: 

   METHOD: 

**********************************************************************/

char *strsave(s)
char *s;

{

/* external functions */

	char	*alloca();
	int	strlen();
	char	*strcpy();

/* local variables */

	static	char	*p;
	static	char	*p_aux;

/* code */

	p_aux = p = alloca((strlen(s) + 1) * sizeof(char));

	for (; *s != '\0'; s++) *p++ = *s;
	*p = '\0';
	return(p_aux);
}


/*************************************************************************

   ROUTINE:     alloca

   FUNCTION:    allocate

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

char *alloca(n_bytes)
int n_bytes;

{
/* external functions */

        char    *malloc();

/* local variables */
	
	static	char *p;

/* code */

	p = malloc(n_bytes);
        if (p == NULL) MEM_OV;
	return(p);
}


/*************************************************************************

   ROUTINE:     release

   FUNCTION:    free space

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

release(addr, n_bytes)
char *addr;
int n_bytes;

{

/* code */

	free(addr);
}




/*************************************************************************

   ROUTINE:	obj_rel()

   FUNCTION:	return a pointer to a release string (without version) 

   AUTHOR(s):	G.Berardo

   DATE:        18 APR 1989

   METHOD:

**********************************************************************/

char *obj_rel(rel_str)
char		*rel_str;	   /* full release string */
{
        char    *strcpy();

  static char obj_rel_str[8];
 
  strcpy(obj_rel_str,rel_str);
  if (isalpha(rel_str[strlen(rel_str)-1]))
    obj_rel_str[strlen(rel_str)-1] = '\0';
  return (obj_rel_str);
}

/*************************************************************************

   ROUTINE:	match_rel()

   FUNCTION:	return a pointer to a release string (without version) 

   AUTHOR(s):	G.Berardo

   DATE:        18 APR 1989

   METHOD:

**********************************************************************/

char *match_rel(rel_str)
char		*rel_str;	   /* full release string */
{
        char    *strcpy();

  static char match_rel_str[8];
 
  strcpy(match_rel_str,rel_str);
  if (isalpha(rel_str[strlen(rel_str)-1]))
    match_rel_str[strlen(rel_str)-1] = '\0';
  return (match_rel_str);
}

