/* LANGUAGE:    C        VAX VMS COMPILER VERSION 2.0-003

 PACKAGE NAME:          LVARI

 PACKAGE ROUTINES:      new_c_list()
                        new_c_el()
			create_conn_block()
                        last_c()
                        enqueue_c()
                        dequeue_c()
			free_list()
			free_conn()
                        first_c()
                        set_queue_c()
                        ins_last()
                        search_el()
                        search_ty()
                        cancel()
                        enqueue_u()
                        enqueue_l()
                        dequeue_u()
                        dequeue_l()
                        push()
                        pop()
                        signal_loop()
                        enqueue()
                        dequeue()
                        enqueue_in()
                        dequeue_in()
                        init_stack()
			ret_in()
			cnt_lst()
			cnt_con_lst()
			cnt_con_mod()

 PACKAGE FUNCTION:      Queues, Stacks, Lists elaborations

 PACKAGE AUTHOR(s):     Daniela Magnani

 PACKAGE CREATION DATE: 19/08/1986

**********************************************************************/

# include <stdio.h>
# include "ldef.h"
# include "lstruct.h"

/***************************************************************************/
/*                                                                         */
/*                          VARIABLES DECLARATION                          */
/*                                                                         */
/***************************************************************************/

unsigned long           n_enqueue_l = 0;
unsigned long           n_enqueue = 0;      /* enqueue amount */

static CONN_PTR         head_conn;
static CONN_PTR         last_conn;
static FIFOL_PTR        p_sl = NULL;  /* rear pointer */                
static FIFOL_PTR        p_el = NULL;  /* front pointer */
static FIFOU_PTR        p_su = NULL;
static FIFOU_PTR        p_eu = NULL;
static ST_PTR		is;	/* stack first element pointer */
static unsigned long	el_st;	/* stack counter */
static QUEUE_PTR        p_s; /* queue first element pointer */
static QUEUE_PTR        p_e; /* queue last element pointer */
static IN_L_PTR         p_si = NULL;
static IN_L_PTR         p_ei = NULL;

/* conn elements allocation */

static CONN_PTR		free_conn_list = NULL;
long			free_conn_cnt = 0;
long			block_conn_cnt = 0;


/*************************************************************************

   ROUTINE:     new_c_list

   FUNCTION:    new CONN list initialization

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

new_c_list()

{
/* code */

        head_conn = NULL;            /* front pointer */
        last_conn = NULL;            /* rear pointer  */
}


/*************************************************************************

   ROUTINE:     new_c_el

   FUNCTION:    new CONN element creation 
		
   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 	Verifica se esistono elementi nella free list.
		Se non esistono crea un blocco di elementi di connectivity
		e lo linka alla free list.

   METHOD:

**********************************************************************/

CONN_PTR new_c_el(type, signal, first, second)
int type;
int signal;
int first;
int second;
{

/* local variables */

        static CONN_PTR        p;

/* code */

	if (free_conn_list == NULL) create_conn_block();

	p = free_conn_list;
	free_conn_list = p->c_next;
	free_conn_cnt--;

	p->signal = signal;
	p->type = type;
	p->CC.first = first;
	p->CC.second = second;
        p->c_next = NULL;
        return(p);
}


/*************************************************************************

   ROUTINE:     last_c

   FUNCTION:    Return the address of the last element of a CONN list

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

CONN_PTR last_c(p)
CONN_PTR p;     /* list start pointer */

{
/* code */

        if (p != NULL)
        {
                while (p->c_next != NULL) p = p->c_next;
        }
        return(p);
}


/*************************************************************************

   ROUTINE:     enqueue_c

   FUNCTION:    insertion of an element into CONN list

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD:      FIFO

**********************************************************************/

enqueue_c(p)
CONN_PTR p;

{
/* code */

        if (last_conn == NULL)
        {
                last_conn = head_conn = p;
        }
        else
        {
                last_conn->c_next = p;
                last_conn = p;
        }
}


/*************************************************************************

   ROUTINE:     dequeue_c

   FUNCTION:    Delete the first element of a CONN list.
                Return this element to the calling function.

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

CONN_PTR dequeue_c()

{
/* local variables */

        CONN_PTR        p = NULL;

/* code */

        if (head_conn != NULL)
        {
                p = head_conn;
                head_conn = head_conn->c_next;
                if (head_conn == NULL) last_conn = NULL;
        }
        return(p);
}


/*************************************************************************

   ROUTINE:     free_list

   FUNCTION:    Free a CONN list (start pointer is given as function parameter)
		
   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 	The input list may be empty
		The space freed is linked to the connectivity free list

   METHOD: 

**********************************************************************/

free_list(p)
CONN_PTR p;

{

/* code */

	if (p != NULL)
	{
		CONN_PTR aux_p = p;
                for(; p->c_next != NULL; p = p->c_next)	free_conn_cnt++;
		free_conn_cnt++;
		p->c_next = free_conn_list;
		free_conn_list = aux_p;
	}
}


/*************************************************************************

   ROUTINE:     free_conn

   FUNCTION:    Free a CONN element (start pointer is given as function 
		parameter)
		
   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 	The input list may be empty
		The space freed is linked to the connectivity free list

   METHOD: 

**********************************************************************/

free_conn(p)
CONN_PTR p;

{

/* code */

	if (p != NULL)
	{
		p->c_next = free_conn_list;
		free_conn_list = p;
		free_conn_cnt++;
	}
}


/*************************************************************************

   ROUTINE:     first_c

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

CONN_PTR first_c()

{

/* code */

        return(head_conn);
}


/*************************************************************************

   ROUTINE:     set_queue_c

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

set_queue_c(p)
CONN_PTR p;

{
/* external functions */

        CONN_PTR        last_c();

/* code */

        head_conn = p;
        last_conn = last_c(p);
}


/*************************************************************************

   ROUTINE:     ins_last

   FUNCTION:    merge two CONN lists

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

CONN_PTR ins_last(dest, source)
CONN_PTR dest;
CONN_PTR source;

{
/* external functions */

        CONN_PTR        first_c();
        CONN_PTR        last_c();

/* code */
        
        set_queue_c(dest);
        enqueue_c(source);
        /* devo ricalcolare last_conn perche' ENQUEUE_C prevede inserimento
           di un solo elemento e non di una lista */
        last_conn = last_c(last_conn);
        return(first_c());
}


/*************************************************************************

   ROUTINE:     search_el

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

CONN_PTR search_el(t_el, list)
unsigned long t_el;
CONN_PTR list;

{
/* local variables */

        CONN_PTR        p = NULL;

/* code */
        
        while ((list != NULL) && (p == NULL))
        {
                if (list->signal == t_el) p = list;
                list = list->c_next;
        }
        return(p);
}


/*************************************************************************

   ROUTINE:     search_ty

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

CONN_PTR search_ty(s_ty, list)
unsigned char s_ty;
CONN_PTR list;

{
/* local variables */

        CONN_PTR        p = NULL;

/* code */
        
        while ((list != NULL) && (p == NULL))
        {
                if (list->type == s_ty) p = list;
                list = list->c_next;
        }
        return(p);
}


/*************************************************************************

   ROUTINE:     cancel

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

CONN_PTR cancel(p, list)
CONN_PTR p;
CONN_PTR list;

{
/* local variables */

        CONN_PTR        aux;

/* code */
        
        if (p == list) list = p->c_next;
        else
        {
                aux = list;
                while (aux->c_next != p) aux = aux->c_next;
                aux->c_next = p->c_next;
        }
        return(list);
}


/*************************************************************************

   ROUTINE:     enqueue_u

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

enqueue_u(s, s1)
char *s;
char *s1;

{
/* external functions */

        char    *alloca();

/* local variables */

        FIFOU_PTR       p_queue;

/* code */

        p_queue = (FIFOU_PTR) alloca(sizeof(*p_queue));

        p_queue->name = s;
        p_queue->father_name = s1;
        p_queue->next = NULL;
        if (p_eu == NULL)
                p_su = p_eu = p_queue;
        else
                p_eu = p_eu->next = p_queue;
}


/*************************************************************************

   ROUTINE:     enqueue_l

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

enqueue_l(s, un)
char *s;
UNIT_PTR un;

{
/* external functions */

        char    *alloca();

/* local variables */

        FIFOL_PTR       p_queue;

/* code */
        
        p_queue = (FIFOL_PTR) alloca(sizeof(*p_queue));

        p_queue->name = s;
        p_queue->unit_index = un;
        p_queue->next = NULL;
        if (p_el == NULL)
                p_sl = p_el = p_queue;
        else
                p_el = p_el->next = p_queue;
        n_enqueue_l++;
}


/*************************************************************************

   ROUTINE:     dequeue_u

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

FIFOU_PTR dequeue_u()

{
/* local variables */

        FIFOU_PTR       p_queue = NULL;

/* code */
        
        if (p_su != NULL)
        {
                p_queue = p_su;
                p_su = p_su->next;
                if (p_su == NULL) p_eu = p_su;
        }
        return(p_queue);
}


/*************************************************************************

   ROUTINE:     dequeue_l

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

FIFOL_PTR dequeue_l()
{

/* local variables */

        FIFOL_PTR       p_queue = NULL;

/* code */
        
        if (p_sl != NULL)
        {
                p_queue = p_sl;
                p_sl = p_sl->next;
                if (p_sl == NULL) p_el = p_sl;
        }
        return(p_queue);
}


/*************************************************************************

   ROUTINE:     init_stack

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        15/04/1987

   NOTES: 

   METHOD: 

**********************************************************************/

init_stack()
{
        is = NULL;
	el_st = 0;
}


/*************************************************************************

   ROUTINE:     push

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 	The stack is organized like a double linked list in 
		order to simplify the print operation

**********************************************************************/

push(t_el)
unsigned long t_el;

{
/* external functions */
	
	char	*alloca();

/* local variables */

	ST_PTR	l_a;

/* code */

	if (++el_st <= DIM_STACK)
	{
		l_a = (ST_PTR) alloca(sizeof(*l_a));

		l_a->el = t_el;
		l_a->next = is;
		l_a->prev = NULL;
		if (is != NULL) is->prev = l_a;
		is = l_a;
	}
}


/*************************************************************************

   ROUTINE:     pop

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

pop()
{

/* local variables */

	ST_PTR	l_a;

/* code */

	if (--el_st < DIM_STACK)
	{
		if (is != NULL)
		{
			l_a = is;
			is = is->next;
			if (is != NULL) is->prev = NULL;
			free(l_a);
		}
        	else
                	prog_bug("loop stack is empty");
	}
}


/*************************************************************************

   ROUTINE:     signal_loop

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

signal_loop(t_el, s)
unsigned long t_el;
char *s;

{

/* external functions */

        char    	*ret_path();
        
/* local variables */

        unsigned long   i;
        char            path[PATHLENGTH + 1];
	ST_PTR		l_a;

/* code */

        for (l_a = is; l_a != NULL && l_a->el != t_el; l_a = l_a->next);
        /* start loop found */
	if (l_a != NULL) start_map(s);
        for (; l_a != NULL; l_a = l_a->prev)
        {
                loop_map_s(ret_path(l_a->el, path));
        }
}


/*************************************************************************

   ROUTINE:     ret_path

   FUNCTION:    return the path name of the input topological element.
                If the name is undefined returns the topological index.

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

char *ret_path(t_el, name)
unsigned long t_el;
char *name;
{

/* external variables */

        extern EHEADER          h_exe;

/* external functions */

        char    *strcat();
        char    *ret_in();
        int     strlen();
	int	strcmp();
        
/* code */
	
        if (H_S[t_el]->s_tab == NULL)
	{
                sprintf(name, "%u", t_el);
		if (t_el == VDD) sprintf(name, "vdd");
		if (t_el == GND) sprintf(name, "gnd");
	}
        else
        {
                ret_in(H_S[t_el]->path_ptr, name);        

                /* aggiungo symbol name */
                if ((strlen(name) + NAMELENGTH + 1) > 
                                        PATHLENGTH) PATH_ERR;
		if (strcmp(name, NULL_NAME) != 0) strcat(name, SEP_STR);
		strcat(name, H_S[t_el]->s_tab->el_name);
        }
        return(name);
}


/*************************************************************************

   ROUTINE:     enqueue

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

enqueue(el)
unsigned long el;
{

/* external functions */

        char    *alloca();

/* local variables */

        QUEUE_PTR       p_queue;

/* code */

        p_queue = (QUEUE_PTR) alloca(sizeof(*p_queue));

        p_queue->el = el;
        p_queue->pun = NULL;
        if (! p_e)
                p_s = p_e = p_queue;
        else
                p_e = p_e->pun = p_queue;
        n_enqueue++;
}


/*************************************************************************

   ROUTINE:     dequeue

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

QUEUE_PTR dequeue()

{
/* local variables */

        QUEUE_PTR       p_queue = NULL;

/* code */

        if (p_s)
        {
                p_queue = p_s;
                p_s = p_s->pun;
                if (!p_s) p_e = p_s;
        }
        return(p_queue);
}


/*************************************************************************

   ROUTINE:     enqueue_in

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

enqueue_in(node, val, unit)
unsigned long node;
unsigned long val;
UNIT_PTR unit;

{
/* external functions */

        char    *alloca();

/* local variables */

        IN_L_PTR        p_queue;

/* code */

        p_queue = (IN_L_PTR) alloca(sizeof(*p_queue));

        p_queue->node = node;
        p_queue->val = val;
        p_queue->unit = unit;
        p_queue->pun = NULL;

        if (! p_ei)
                p_si = p_ei = p_queue;
        else
                p_ei = p_ei->pun = p_queue;
}


/*************************************************************************

   ROUTINE:     dequeue_in

   FUNCTION:    #

   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 

   METHOD: 

**********************************************************************/

IN_L_PTR dequeue_in()

{
/* local variables */

        IN_L_PTR        p_queue = NULL;

/* code */

        if (p_si)
        {
                p_queue = p_si;
                p_si = p_si->pun;
                if (!p_si) p_ei = p_si;
        }
        return(p_queue);
}


/*************************************************************************

   ROUTINE:     ret_in

   FUNCTION:    return the path name associated with an index of the 
		instantiation tree.

   AUTHOR(s):   Daniela Magnani

   DATE:        04/05/1987

   NOTES: 

   METHOD: 

**********************************************************************/

char *ret_in(ind, name)
unsigned long ind;
char *name;
{

/* external variables */

        extern EHEADER          h_exe;
        extern IN_T_PTR         in_tree;

/* external functions */

        char    *strcat();
        char    *alloca();
        char    *strcpy();
        int     strlen();
        
/* local variables */

        unsigned long   i;
        long            *l_i;
        long            *l_aux;

/* code */
	
	l_aux = l_i = (long*) alloca(h_exe.n_inst * sizeof(long));
        
        /* ricostruisco a ritroso indici path name */
        while (ind != MAX_UNSIGNED)
        {
        	*l_i++ = ind;
                ind = in_tree[ind].father;
        }
        /* l_i-- punta all'elemento a livello piu' basso della gerarchia (main)
	l_aux punta a quello a livello piu' alto */
/*        sprintf(name, "%s%s", SEP_STR, in_tree[*(--l_i)].in_name); */
	strcpy(name, NULL_NAME);
	l_i--;

        if (l_i != l_aux)
	{
		if ((strlen(name) + NAMELENGTH + 1) > PATHLENGTH) PATH_ERR;
                strcat(name, in_tree[*(--l_i)].in_name);

	        while (l_i != l_aux)
        	{
		    if ((strlen(name) + NAMELENGTH + 1) > PATHLENGTH) PATH_ERR;
                    strcat(name, SEP_STR);
                    strcat(name, in_tree[*(--l_i)].in_name);
	        }
	}
	free(l_aux);
	return(name);
}


/*************************************************************************

   ROUTINE:     cnt_lst

   FUNCTION:    Count the number of elements into a connectivity list
		(fanin, fanout, conn, code)

   AUTHOR(s):   Daniela Magnani

   DATE:        27/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

unsigned long cnt_lst(ele, p_s, type)
unsigned long ele;
STAT_PTR *p_s;
char type;	/* FANOUT, FANIN, CODE, LCONN */

{

/* external functions */

	unsigned long	cnt_con_lst();

/* local variables */

	CONN_PTR	aux_lst;
        int             partial_count;
/* body */

	
	switch(p_s[ele]->code)
        {
                case NOD :
                case WIR :
                case REG :
			switch (type)
			{
				case FANIN :
					aux_lst = p_s[ele]->C.l_fanin;
					break;
				case FANOUT :
					aux_lst = p_s[ele]->C.l_fanout;
					break;
				case LCONN :
					aux_lst = p_s[ele]->C.l_conn;
					break;
				default :
				 prog_bug("Illegal el type called in cnt_lst");
					break;
			}
                        break;
		case GAT :
			switch (type)
			{
				case FANIN :
					aux_lst = p_s[ele]->G.l_fanin;
					break;
				case FANOUT :
					aux_lst = p_s[ele]->G.l_fanout;
					break;
				default :
				 prog_bug("Illegal el type called in cnt_lst");
					break;
			}
                        break;
		case RT_CN :
		case RT_CNT :
			switch (type)
			{
				case FANIN :
					aux_lst = p_s[ele]->O.l_fanin;
					break;
				case FANOUT :
					aux_lst = p_s[ele]->O.l_fanout;
					break;
				case CODE :
					aux_lst = p_s[ele]->O.l_code;
					break;
				default :
				 prog_bug("Illegal el type called in cnt_lst");
					break;
			}
                        break;
		case RT_AS :
		case RT_LD :
		case BIT_EX :
		case AS_TR :
			switch (type)
			{
				case FANIN :
					aux_lst = p_s[ele]->A.l_fanin;
					break;
				case FANOUT :
					aux_lst = p_s[ele]->A.l_fanout;
					break;
				case CODE :
					aux_lst = p_s[ele]->A.l_code;
					break;
				default :
				 prog_bug("Illegal el type called in cnt_lst");
					break;
			}
                        break;
		case PROC :
			switch (type)
			{
				case FANIN :
					aux_lst = p_s[ele]->P.l_tfanin;
                                        partial_count=cnt_con_lst(aux_lst);
					aux_lst = p_s[ele]->P.l_dfanin;
                                        return(partial_count+cnt_con_lst(aux_lst));
					break;
				case FANOUT :
					aux_lst = p_s[ele]->P.l_fanout;
					break;
				case CODE :
					aux_lst = p_s[ele]->P.l_code;
					break;
				default :
				 prog_bug("Illegal el type called in cnt_lst");
					break;
			}
                        break;
                case CONST :
		  switch (p_s[ele]->I.type)
		  {
		     case C_WIRE :
			switch (type)
			{
				case FANOUT :
				  aux_lst = p_s[ele]->IW.l_fanout;
				break;
				case LCONN :
				  aux_lst = p_s[ele]->IW.l_conn;
				break;
				default :
				  prog_bug("Illegal el type called in cnt_lst");
				break;
			}
                        break;
		     case C_INT :
		     case D_CONST :
			switch (type)
			{
				case FANOUT :
				  aux_lst = p_s[ele]->ID.l_fanout;
				break;
				case FANIN :
				  aux_lst = p_s[ele]->ID.l_fanin;
				break;
				case CODE :
				  aux_lst = p_s[ele]->ID.l_code;
				break;
				default :
				  prog_bug("Illegal el type called in cnt_lst");
				break;
			}
                        break;
		     case C_CHAR :
			switch (type)
			{
				case FANOUT :
				  aux_lst = p_s[ele]->IC.l_fanout;
				break;
				default :
				  prog_bug("Illegal el type called in cnt_lst");
				break;
			}
		  }
                case MEM :
			switch (type)
			{
				case FANOUT :
					aux_lst = p_s[ele]->M.l_fanout;
					break;
				case FANIN :
					aux_lst = p_s[ele]->M.l_fanin;
					break;
				default :
				 prog_bug("Illegal el type called in cnt_lst");
					break;
			}
                        break;
		default :
			prog_bug("Illegal element called in cnt_lst");
			break;
	}

	return(cnt_con_lst(aux_lst));
}


/*************************************************************************

   ROUTINE:     cnt_con_lst

   FUNCTION:    Count the number of elements into a connectivity list

   AUTHOR(s):   Daniela Magnani

   DATE:        27/07/1987

   NOTES: 

   METHOD: 

**********************************************************************/

unsigned long cnt_con_lst(p)
CONN_PTR p;

{

/* local variables */

	static unsigned long	count;

/* body */

	for (count = 0; p != NULL; count++, p = p->c_next);
	return(count);
}

/*************************************************************************

   ROUTINE:     cnt_con_mod

   FUNCTION:    Count the number of elements into a connectivity fanin
                list and modify  second field of the elements connected 
                to AS_TR blocks.

   AUTHOR(s):   Giulio Berardo

   DATE:        17/5/89

   NOTES: 

   METHOD: 

**********************************************************************/

unsigned long cnt_con_mod(p)
CONN_PTR p;

{

/* external variables */
        extern  EHEADER  h_exe;

/* local variables */

	static unsigned long	count;

/* body */

	for (count = 0; p != NULL; count++, p = p->c_next)
        {
          if (H_S[p->signal]->code == AS_TR) p->CC.second = C_AS_TR;
        }
	return(count);
}



/*************************************************************************

   ROUTINE:     create_conn_block

   FUNCTION:    create a block of CONN elements
		
   AUTHOR(s):   Daniela Magnani

   DATE:        22/08/1986

   NOTES: 	The block is linked to the free conn list.
		The block size is 13 * 512 = 6656

   METHOD:

**********************************************************************/

create_conn_block()
{


/* external functions */

	char	*alloca();

/* local variables */

	char	*p;
	int	i;

/* code */

	p = alloca(NCON * DIMCON);
	for (i = 0; i < ((NCON - 1) * DIMCON); i += DIMCON)
		((CONN_PTR)((unsigned char *)p + i))->c_next = 
		 (CONN_PTR)((unsigned char *)p + i + DIMCON);

	((CONN_PTR)((unsigned char *)p + i))->c_next = free_conn_list;
	free_conn_cnt += NCON;
	block_conn_cnt++;
	free_conn_list = (CONN_PTR)p;
}

